#    builder.py -- Classes for building packages
#    Copyright (C) 2006, 2007 James Westby <jw+debian@jameswestby.net>
#
#    This file is part of breezy-debian.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from __future__ import absolute_import

import shutil
import subprocess
import os
import tempfile

from ...errors import BzrError
from ...trace import note

from .hooks import run_hook
from .util import (
    get_parent_dir,
    subprocess_setup,
    changes_filename,
    dget_changes,
    )


class ChangesFileMissing(BzrError):

    _fmt = "Missing changes file: %(path)r"


class NoSourceDirError(BzrError):

    _fmt = ("There is no existing source directory to use. Use "
            "--export-only or --dont-purge to get one that can be used")


class BuildFailedError(BzrError):
    _fmt = "The build failed."


class DebBuild(object):
    """The object that does the building work."""

    def __init__(self, distiller, target_dir, builder, use_existing=False):
        """Create a builder.

        :param distiller: the SourceDistiller that will get the source to
            build.
        :param target_dir: the directory in which to do all the work.
        :param builder: the build command to use.
        :param use_existing: whether to re-use the target_dir if it exists.
        """
        self.distiller = distiller
        self.target_dir = target_dir
        self.builder = builder
        self.use_existing = use_existing

    def prepare(self):
        """Do any preparatory steps that should be run before the build.

        It checks that everything is well, and that some needed dirs are
        created.
        """
        parent_dir = get_parent_dir(self.target_dir)
        if parent_dir != '' and not os.path.exists(parent_dir):
            os.makedirs(parent_dir)
        if os.path.exists(self.target_dir):
            if not self.use_existing:
                note("Purging the build dir: %s", self.target_dir)
                shutil.rmtree(self.target_dir)
            else:
                note("Not purging build dir as requested: %s",
                     self.target_dir)
        else:
            if self.use_existing:
                raise NoSourceDirError

    def export(self):
        self.distiller.distill(self.target_dir)

    def before_build(self):
        subprocess.check_call(
            ['dpkg-source', '--before-build', self.target_dir],
            preexec_fn=subprocess_setup)

    def after_build(self):
        subprocess.check_call(
            ['dpkg-source', '--after-build', self.target_dir],
            preexec_fn=subprocess_setup, )

    def build(self):
        """This builds the package using the supplied command."""
        note("Building the package in %s, using %s", self.target_dir,
             self.builder)
        proc = subprocess.Popen(
            self.builder, shell=True, cwd=self.target_dir,
            preexec_fn=subprocess_setup)
        proc.wait()
        if proc.returncode != 0:
            raise BuildFailedError

    def clean(self):
        """This removes the build directory."""
        note("Cleaning build dir: %s", self.target_dir)
        shutil.rmtree(self.target_dir)


def do_build(package_name, version, distiller, local_tree, config,
             build_command, target_dir=None):
    """Actually run a build."""
    with tempfile.TemporaryDirectory() as bd:
        build_source_dir = os.path.join(
            bd, package_name + "-" + version.upstream_version)
        builder = DebBuild(
                distiller, build_source_dir,
                build_command,
                use_existing=False)
        builder.prepare()
        run_hook(local_tree, 'pre-export', config)
        builder.export()
        run_hook(local_tree, 'pre-build', config, wd=build_source_dir)
        builder.build()
        run_hook(local_tree, 'post-build', config, wd=build_source_dir)
        changes = changes_filename(
            package_name, version, 'source')
        changes_path = os.path.join(bd, changes)
        if target_dir is not None:
            if not os.path.exists(changes_path):
                raise ChangesFileMissing(changes_path)
            return dget_changes(changes_path, target_dir)



# This file is a modified copy of bzrtools' upstream_import.py, last changed in
# bzrtools 1.14.0.

"""Import upstream source into a branch"""

from __future__ import absolute_import

from io import BytesIO

from ...upstream_import import (
    add_implied_parents,
    common_directory,
    do_directory,
    names_of_files,
    DirWrapper,
    )
from ...bzr import generate_ids

from contextlib import ExitStack

from ... import urlutils
from ...controldir import ControlDir
from ...errors import NoSuchFile, BzrCommandError, NotBranchError, BzrError
from ...osutils import (
    file_iterator,
    basename,
    file_kind,
    splitpath,
    normpath,
    is_inside_any,
    )
from ...trace import warning
from ...tree import Tree
from ...transform import resolve_conflicts


class UnknownType(BzrError):
    _fmt = 'Cannot extract "%(path)s" from archive as it is an unknown type.'

    def __init__(self, path):
        BzrError.__init__(self, path=path)


files_to_ignore = set(
    ['.shelf', '.bzr', '.bzr.backup', '.bzrtags',
     '.bzr-builddeb'])


def should_ignore(relative_path: str) -> bool:
    parts = splitpath(relative_path)
    if not parts:
        return False
    for part in parts:
        if part in files_to_ignore:
            return True
        if part.endswith(',v'):
            return True


def import_dir(
        tree: Tree, dir: str, file_ids_from=None, target_tree=None,
        exclude=None):
    dir_input = BytesIO(dir.encode('utf-8'))
    dir_input.seek(0)
    dir_file = DirWrapper(dir_input)
    if file_ids_from is None:
        file_ids_from = []
    with ExitStack() as es:
        for other_tree in file_ids_from:
            es.enter_context(other_tree.lock_read())
        return _import_archive(
            tree, dir_file, file_ids_from, target_tree=target_tree,
            exclude=exclude)


def _get_paths_to_process(
        archive_file: str, prefix,
        implied_parents, exclude=None):
    to_process = set()
    for member in archive_file.getmembers():
        if member.type == 'g':
            # type 'g' is a header
            continue
        relative_path = member.name
        relative_path = normpath(relative_path)
        relative_path = relative_path.lstrip('/')
        if prefix is not None:
            relative_path = relative_path[len(prefix)+1:]
            relative_path = relative_path.rstrip('/')
        if relative_path == '' or relative_path == '.':
            continue
        if should_ignore(relative_path):
            continue
        if exclude and is_inside_any(exclude, relative_path):
            continue
        add_implied_parents(implied_parents, relative_path)
        to_process.add((relative_path, member))
    return to_process


def _import_archive(
        tree: Tree, archive_file: str, file_ids_from,
        target_tree = None, exclude=None):
    prefix = common_directory(names_of_files(archive_file))
    try:
        transform = tree.transform
    except AttributeError:  # brz < 3.1.1
        transform = tree.get_transform
    with transform() as tt:
        removed = set()
        for path, entry in tree.iter_entries_by_dir():
            if entry.parent_id is None:
                continue
            trans_id = tt.trans_id_tree_path(path)
            tt.delete_contents(trans_id)
            removed.add(path)

        added = set()
        implied_parents = set()
        seen = set()
        to_process = _get_paths_to_process(
            archive_file, prefix, implied_parents, exclude=exclude)
        renames = {}

        if not tree.supports_setting_file_ids():
            file_ids_from = []

        # First we find the renames
        other_trees = file_ids_from[:]
        if target_tree is not None and tree.supports_setting_file_ids():
            other_trees.insert(0, target_tree)
        for other_tree in other_trees:
            for relative_path, member in to_process:
                trans_id = tt.trans_id_tree_path(relative_path)
                existing_file_id = tt.tree_file_id(trans_id)
                target_id = other_tree.path2id(relative_path)
                if (target_id is not None
                        and target_id != existing_file_id
                        and target_id not in renames):
                    renames[target_id] = relative_path

        # The we do the work
        for relative_path, member in to_process:
            trans_id = tt.trans_id_tree_path(relative_path)
            added.add(relative_path.rstrip('/'))
            # To handle renames, we need to not use the preserved file id,
            # rather we need to lookup the file id in target_tree, if there is
            # one. If there isn't, we should use the one in the current tree,
            # and failing that we will allocate one. In this importer we want
            # the target_tree to be authoritative about id2path, which is why
            # we consult it first.
            if tree.supports_setting_file_ids():
                existing_file_id = tt.tree_file_id(trans_id)
                # If we find an id that we know we are going to assign to
                # different path as it has been renamed in one of the
                # file_ids_from trees then we ignore the one in this tree.
                if existing_file_id in renames:
                    if relative_path != renames[existing_file_id]:
                        existing_file_id = None
                found_file_id = None
                if target_tree is not None:
                    found_file_id = target_tree.path2id(relative_path)
                    if found_file_id in renames:
                        if renames[found_file_id] != relative_path:
                            found_file_id = None
                if found_file_id is None and existing_file_id is None:
                    for other_tree in file_ids_from:
                        found_file_id = other_tree.path2id(relative_path)
                        if found_file_id is not None:
                            if found_file_id in renames:
                                if renames[found_file_id] != relative_path:
                                    found_file_id = None
                                    continue
                            break
                if found_file_id is not None and found_file_id != existing_file_id:
                    # Found a specific file id in one of the source trees
                    tt.version_file(trans_id=trans_id, file_id=found_file_id)
                    if existing_file_id is not None:
                        # We need to remove the existing file so it can be
                        # replaced by the file (and file id) from the
                        # file_ids_from tree.
                        tt.delete_versioned(trans_id)
                    trans_id = tt.trans_id_file_id(found_file_id)
                if not found_file_id and not existing_file_id:
                    # No file_id in any of the source trees and no file id in
                    # the base tree.
                    name = basename(member.name.rstrip('/'))
                    file_id = generate_ids.gen_file_id(name)
                    tt.version_file(file_id=file_id, trans_id=trans_id)
            else:
                try:
                    tt.version_file(trans_id=trans_id)
                except TypeError:  # brz < 3.1.1
                    name = basename(member.name.rstrip('/'))
                    file_id = generate_ids.gen_file_id(name)
                    tt.version_file(trans_id=trans_id, file_id=file_id)

            path = tree.abspath(relative_path)
            if member.name in seen:
                if tt.final_kind(trans_id) == 'file':
                    tt.set_executability(None, trans_id)
                tt.cancel_creation(trans_id)
            seen.add(member.name)
            if member.isreg():
                tt.create_file(file_iterator(archive_file.extractfile(member)),
                               trans_id)
                executable = (member.mode & 0o111) != 0
                tt.set_executability(executable, trans_id)
            elif member.isdir():
                do_directory(tt, trans_id, tree, relative_path, path)
            elif member.issym():
                tt.create_symlink(member.linkname, trans_id)
            else:
                raise UnknownType(relative_path)

        for relative_path in implied_parents.difference(added):
            if relative_path == "":
                continue
            trans_id = tt.trans_id_tree_path(relative_path)
            path = tree.abspath(relative_path)
            do_directory(tt, trans_id, tree, relative_path, path)
            if tt.tree_file_id(trans_id) is None:
                found = False
                for other_tree in file_ids_from:
                    with other_tree.lock_read():
                        if other_tree.has_filename(relative_path):
                            file_id = other_tree.path2id(relative_path)
                            if file_id is not None:
                                tt.version_file(
                                    trans_id=trans_id, file_id=file_id)
                                found = True
                                break
                if not found:
                    # Should this really use the trans_id as the
                    # file_id?
                    tt.version_file(trans_id=trans_id, file_id=trans_id)
            added.add(relative_path)

        for path in removed.difference(added):
            tt.unversion_file(tt.trans_id_tree_path(path))

        if getattr(tt, 'cook_conflicts', None):
            conflicts = tt.cook_conflicts(resolve_conflicts(tt))
        else:  # brz < 3.1.1
            from breezy.transform import cook_conflicts
            conflicts = cook_conflicts(resolve_conflicts(tt), tt)

        for conflict in conflicts:
            warning('%s', conflict)
        tt.apply()



#    test_dep3.py -- Blackbox tests for dep3-patch.
#    Copyright (C) 2011 Canonical Ltd.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#


"""Blackbox tests for "bzr dep3-patch"."""

from __future__ import absolute_import

from .....tests.blackbox import ExternalBase

import os


class TestDep3Patch(ExternalBase):

    def setUp(self):
        super(TestDep3Patch, self).setUp()
        self.upstream_tree = self.make_branch_and_tree("upstream")
        self.upstream_tree.commit(message="initial commit")
        packaging = self.upstream_tree.controldir.sprout("packaging")
        self.packaging_tree = packaging.open_workingtree()
        feature = self.upstream_tree.controldir.sprout("feature")
        self.feature_tree = feature.open_workingtree()

    def test_nothing_to_do(self):
        (out, err) = self.run_bzr("dep3-patch -d packaging feature", retcode=3)
        self.assertEquals("brz: ERROR: No unmerged revisions\n", err)
        self.assertEquals("", out)

    def test_simple(self):
        # If there is a single revision the commit message from
        # that revision will be used.
        self.build_tree_contents([("feature/foo", "bar\n")])
        self.feature_tree.add("foo")
        self.feature_tree.commit(message="A message", timestamp=1304850124,
            timezone=0, authors=["Jelmer <jelmer@debian.org>"], rev_id=b"therevid")
        (out, err) = self.run_bzr("dep3-patch -d packaging feature")
        self.assertEqualDiff(out, "Description: A message\n"
            "Origin: commit, revision id: therevid\n"
            "Author: Jelmer <jelmer@debian.org>\n"
            "Last-Update: 2011-05-08\n"
            "X-Bzr-Revision-Id: therevid\n"
            "\n"
            "=== added file 'foo'\n"
            "--- old/foo\t1970-01-01 00:00:00 +0000\n"
            "+++ new/foo\t2011-05-08 10:22:04 +0000\n"
            "@@ -0,0 +1,1 @@\n"
            "+bar\n"
            "\n")

    def test_uses_single_revision_commit(self):
        # If there is a single revision the commit message from
        # that revision will be used.
        self.feature_tree.commit(message="A message", timestamp=1304850124,
            timezone=0, authors=["Jelmer <jelmer@debian.org>"])
        (out, err) = self.run_bzr("dep3-patch -d packaging feature")
        self.assertContainsRe(out, "Description: A message\n")

    def test_uses_config_description(self):
        config = self.feature_tree.branch.get_config()
        config.set_user_option("description", "What this does")
        self.feature_tree.commit(message="a message")
        (out, err) = self.run_bzr("dep3-patch -d packaging feature")
        self.assertContainsRe(out, "Description: What this does\n")

    def test_upstream_branch(self):
        os.mkdir('packaging/.bzr-builddeb/')
        with open('packaging/.bzr-builddeb/local.conf', 'w') as f:
            f.write('[BUILDDEB]\nupstream-branch = %s\n' %
                    self.upstream_tree.branch.base)
        self.feature_tree.commit(message="a message")
        (out, err) = self.run_bzr("dep3-patch -d packaging feature")
        self.assertContainsRe(out, "Applied-Upstream: no\n")

    def test_upstream_branch_disabled(self):
        os.mkdir('packaging/.bzr-builddeb/')
        with open('packaging/.bzr-builddeb/local.conf', 'w') as f:
            f.write('[BUILDDEB]\nupstream-branch = %s\n' %
                    self.upstream_tree.branch.base)
        self.feature_tree.commit(message="a message")
        (out, err) = self.run_bzr("dep3-patch --no-upstream-check -d packaging feature")
        self.assertNotContainsRe(out, "Applied-Upstream")

    def test_range(self):
        # A range of revisions can be specified.
        self.build_tree_contents([("feature/foo", "bar\n")])
        self.feature_tree.add("foo")
        self.feature_tree.commit(message="Another message", timestamp=1304850124,
            timezone=0, authors=["Jelmer <jelmer@debian.org>"], rev_id=b"baserevid")
        self.build_tree_contents([("feature/foo", "bla\n")])
        self.feature_tree.commit(message="A message", timestamp=1304850124,
            timezone=0, authors=["Jelmer <jelmer@debian.org>"], rev_id=b"therevid")
        (out, err) = self.run_bzr("dep3-patch -c -1 -d packaging feature")
        self.assertEqualDiff(out, "Description: A message\n"
            "Origin: commit, revision id: therevid\n"
            "Author: Jelmer <jelmer@debian.org>\n"
            "Last-Update: 2011-05-08\n"
            "X-Bzr-Revision-Id: therevid\n"
            "\n"
            "=== modified file 'foo'\n"
            "--- old/foo\t2011-05-08 10:22:04 +0000\n"
            "+++ new/foo\t2011-05-08 10:22:04 +0000\n"
            "@@ -1,1 +1,1 @@\n"
            "-bar\n"
            "+bla\n"
            "\n")

    def test_open_ended_range(self):
        # If there is a single revision the commit message from
        # that revision will be used.
        self.feature_tree.commit(message="A message", timestamp=1304850124,
            timezone=0, authors=["Jelmer <jelmer@debian.org>"])
        (out, err) = self.run_bzr("dep3-patch -d packaging feature -r-2..")
        self.assertContainsRe(out, "Description: A message\n")



#    source_distiller.py -- Getting the source to build from a branch
#    Copyright (C) 2008, 2009 Canonical Ltd.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from __future__ import absolute_import

from debian.changelog import Changelog

import os
import shutil
import subprocess
import sys
import tempfile

from ... import errors as bzr_errors
from ...export import (
    export,
    )

from .util import (
    extract_orig_tarballs,
    get_parent_dir,
    recursive_copy,
    )


class SourceDistiller(object):
    """A source distiller extracts the source to build from a location.

    It does whatever is needed to give you a source you can build at
    a location of your choice.
    """

    def __init__(self, tree, subpath):
        """Create a SourceDistiller to distill from the specified tree.

        :param tree: The tree to use as the source.
        :param subpath: subpath in the tree where the package lives
        """
        self.tree = tree
        self.subpath = subpath

    def distill(self, target):
        """Extract the source to a tree rooted at the given location.

        The passed location cannot already exist. If it does then
        FileExists will be raised.

        :param target: a string containing the location at which to
            place the tree containing the buildable source.
        """
        raise NotImplementedError(self.distill)


class NativeSourceDistiller(SourceDistiller):
    """A SourceDistiller for unpacking a native package from a branch."""

    def __init__(self, tree, subpath):
        """Create a SourceDistiller to distill from the specified tree.

        :param tree: The tree to use as the source.
        :param subpath: subpath in the tree where the package lives
        """
        super(NativeSourceDistiller, self).__init__(tree, subpath)

    def distill(self, target):
        """Extract the source to a tree rooted at the given location.

        The passed location cannot already exist. If it does then
        FileExists will be raised.

        :param target: a string containing the location at which to
            place the tree containing the buildable source.
        """
        if os.path.exists(target):
            raise bzr_errors.FileExists(target)
        export(self.tree, target, subdir=self.subpath)


class FullSourceDistiller(SourceDistiller):
    """A SourceDistiller for full-source branches, a.k.a. normal mode"""

    def __init__(self, tree, subpath, upstream_provider):
        """Create a SourceDistiller to distill from the specified tree.

        :param tree: The tree to use as the source.
        :param subpath: subpath in the tree where the package lives
        :param upstream_provider: an UpstreamProvider to provide the upstream
            tarball if needed.
        """
        super(FullSourceDistiller, self).__init__(tree, subpath)
        self.upstream_provider = upstream_provider

    def distill(self, target):
        """Extract the source to a tree rooted at the given location.

        The passed location cannot already exist. If it does then
        FileExists will be raised.

        :param target: a string containing the location at which to
            place the tree containing the buildable source.
        """
        if os.path.exists(target):
            raise bzr_errors.FileExists(target)
        parent_dir = get_parent_dir(target)
        self.upstream_provider.provide(parent_dir)
        export(self.tree, target, subdir=self.subpath)
        # TODO(jelmer): Unapply patches, if they're applied.


class MergeModeDistiller(SourceDistiller):

    def __init__(self, tree, subpath, upstream_provider, top_level=False,
                 use_existing=False):
        """Create a SourceDistiller to distill from the specified tree.

        :param tree: The tree to use as the source.
        :param subpath: subpath in the tree where the package lives
        :param upstream_provider: an UpstreamProvider to provide the upstream
            tarball if needed.
        :param top_level: if the tree is in the top level directory instead of
            inside debian/.
        :param use_existing: whether the distiller should re-use an existing
            target if the distiller supports it.
        """
        super(MergeModeDistiller, self).__init__(tree, subpath)
        self.upstream_provider = upstream_provider
        self.top_level = top_level
        self.use_existing = use_existing

    def distill(self, target):
        """Extract the source to a tree rooted at the given location.

        The passed location cannot already exist. If it does then
        FileExists will be raised.

        :param target: a string containing the location at which to
            place the tree containing the buildable source.
        """
        if not self.use_existing:
            if os.path.exists(target):
                raise bzr_errors.FileExists(target)
        elif self.use_existing:
            if not os.path.exists(target):
                raise bzr_errors.NotADirectory(target)

        # Get the upstream tarball
        parent_dir = get_parent_dir(target)
        if parent_dir != '' and not os.path.exists(parent_dir):
            os.makedirs(parent_dir)
        if not self.use_existing:
            tarballs = self.upstream_provider.provide(parent_dir)
            if not os.path.exists(target):
                os.mkdir(target)
            extract_orig_tarballs(tarballs, target)
        # Now export the tree to provide the debian dir
        with tempfile.TemporaryDirectory(
                prefix='builddeb-merge-debian-') as basetempdir:
            tempdir = os.path.join(basetempdir, "export")
            if self.top_level:
                os.makedirs(tempdir)
                export_dir = os.path.join(tempdir, 'debian')
            else:
                export_dir = tempdir
            export(self.tree, export_dir, subdir=self.subpath)
            # Remove any upstream debian dir, or from previous export with
            # use_existing
            if os.path.exists(os.path.join(target, 'debian')):
                shutil.rmtree(os.path.join(target, 'debian'))
            if sys.version_info >= (3, 8):
                shutil.copytree(
                    tempdir, target, symlinks=True, dirs_exist_ok=True)
            else:
                recursive_copy(tempdir, target)


class DebcargoError(bzr_errors.BzrError):

    _fmt = "Debcargo failed to run."


class DebcargoDistiller(SourceDistiller):
    """A SourceDistiller for unpacking a debcargo package."""

    def __init__(self, tree, subpath, top_level=False, use_existing=False):
        """Create a SourceDistiller to distill from the specified tree.

        :param tree: The tree to use as the source.
        :param subpath: subpath in the tree where the package lives
        """
        super(DebcargoDistiller, self).__init__(tree, subpath)
        self.top_level = top_level
        self.use_existing = use_existing

    def distill(self, target):
        """Extract the source to a tree rooted at the given location.

        The passed location cannot already exist. If it does then
        FileExists will be raised.

        :param target: a string containing the location at which to
            place the tree containing the buildable source.
        """
        if os.path.exists(target):
            raise bzr_errors.FileExists(target)
        with self.tree.get_file(os.path.join(
                self.subpath,
                'debian/changelog' if not self.top_level else
                'changelog'), 'r') as f:
            cl = Changelog(f, max_blocks=1)
            package = cl.package

        if not package.startswith('rust-'):
            raise NotImplementedError

        crate = package[len('rust-'):]

        debcargo_path = [self.subpath]
        if not self.top_level:
            debcargo_path.append('debian')
        debcargo_path.append('debcargo.toml')
        try:
            subprocess.check_call([
                'debcargo', 'package',
                '--changelog-ready',
                '--config', self.tree.abspath(os.path.join(*debcargo_path)),
                '--directory', target, crate])
        except subprocess.CalledProcessError:
            raise DebcargoError()



#    __init__.py -- Testsuite for builddeb
#    Copyright (C) 2007 James Westby <jw+debian@jameswestby.net>
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from __future__ import absolute_import

import shutil
import subprocess
import sys
import tarfile
import zipfile

import doctest
import os

from .... import tests

from debian.changelog import Version, Changelog

from ....tests import (  # noqa: F401
    multiply_tests,
    TestCaseWithTransport,
    TestCaseInTempDir,
    )
from ....tests.features import (  # noqa: F401
    ExecutableFeature,
    ModuleAvailableFeature,
    UnicodeFilenameFeature,
    )


def make_new_upstream_dir(source, dest):
    shutil.copytree(source, dest)


def make_new_upstream_tarball(source, dest):
    tar = tarfile.open(dest, 'w:gz')
    try:
        tar.add(source)
    finally:
        tar.close()


def make_new_upstream_tarball_bz2(source, dest):
    tar = tarfile.open(dest, 'w:bz2')
    try:
        tar.add(source)
    finally:
        tar.close()


def make_new_upstream_tarball_zip(source, dest):
    zip = zipfile.ZipFile(dest, 'w')
    try:
        zip.writestr(source, '')
        for (dirpath, dirnames, names) in os.walk(source):
            for dir in dirnames:
                zip.writestr(os.path.join(dirpath, dir, ''), '')
            for name in names:
                zip.write(os.path.join(dirpath, name))
    finally:
        zip.close()


def make_new_upstream_tarball_bare(source, dest):
    tar = tarfile.open(dest, 'w')
    try:
        tar.add(source)
    finally:
        tar.close()


def make_new_upstream_tarball_xz(source, dest):
    uncompressed = dest + ".temp"
    tar = tarfile.open(uncompressed, 'w')
    try:
        tar.add(source)
    finally:
        tar.close()
    subprocess.check_call(["xz", "-z", uncompressed])
    os.rename(uncompressed+".xz", dest)


def make_new_upstream_tarball_lzma(source, dest):
    import lzma
    f = lzma.LZMAFile(dest, 'w', format=lzma.FORMAT_ALONE)
    try:
        tar = tarfile.open(None, 'w', f)
        try:
            tar.add(source)
        finally:
            tar.close()
    finally:
        f.close()


def load_tests(loader, basic_tests, pattern):
    testmod_names = [
            'blackbox',
            'test_builder',
            'test_bzrtools_import',
            'test_commit_message',
            'test_config',
            'test_dep3',
            'test_directory',
            'test_extract',
            'test_hooks',
            'test_import_dsc',
            'test_merge_changelog',
            'test_merge_package',
            'test_merge_upstream',
            'test_repack_tarball_extra',
            'test_revspec',
            'test_source_distiller',
            'test_upstream',
            'test_upstream_uscan',
            'test_util',
            'test_tagging',
            ]
    basic_tests.addTest(loader.loadTestsFromModuleNames(
        ["%s.%s" % (__name__, i) for i in testmod_names]))

    doctest_mod_names = [
             'config'
             ]
    for mod in doctest_mod_names:
        basic_tests.addTest(
            doctest.DocTestSuite("breezy.plugins.debian." + mod))
    repack_tarball_tests = loader.loadTestsFromModuleNames(
            ['%s.test_repack_tarball' % __name__])
    scenarios = [('dir', dict(build_tarball=make_new_upstream_dir,
                              old_tarball='../package-0.2')),
                 ('.tar.gz', dict(build_tarball=make_new_upstream_tarball,
                                  old_tarball='../package-0.2.tar.gz')),
                 ('.tar.bz2', dict(build_tarball=make_new_upstream_tarball_bz2,
                                   old_tarball='../package-0.2.tar.bz2')),
                 ('.tar.xz', dict(build_tarball=make_new_upstream_tarball_xz,
                                  old_tarball='../package-0.2.tar.xz',
                                  _test_needs_features=[XzFeature])),
                 ('.tar.lzma', dict(
                     build_tarball=make_new_upstream_tarball_lzma,
                     old_tarball='../package-0.2.tar.lzma',
                     _test_needs_features=[LzmaFeature])),
                 ('.zip', dict(build_tarball=make_new_upstream_tarball_zip,
                               old_tarball='../package-0.2.zip')),
                 ('.tar', dict(build_tarball=make_new_upstream_tarball_bare,
                               old_tarball='../package-0.2.tar')),
                 ]
    basic_tests = multiply_tests(repack_tarball_tests, scenarios, basic_tests)
    return basic_tests


class BuilddebTestCase(tests.TestCaseWithTransport):

    package_name = 'test'
    package_version = Version('0.1-1')
    upstream_version = property(
        lambda self: self.package_version.upstream_version)

    def make_changelog(self, version=None):
        if version is None:
            version = self.package_version
        c = Changelog()
        c.new_block()
        c.version = Version(version)
        c.package = self.package_name
        c.distributions = 'unstable'
        c.urgency = 'low'
        c.author = 'James Westby <jw+debian@jameswestby.net>'
        c.date = 'Thu,  3 Aug 2006 19:16:22 +0100'
        c.add_change('')
        c.add_change('  *  test build')
        c.add_change('')
        return c

    def write_changelog(self, changelog, filename):
        with open(filename, 'w') as f:
            changelog.write_to_open_file(f)

    def check_tarball_contents(self, tarball, expected, basedir=None,
                               skip_basedir=False, mode=None):
        """Test that the tarball has certain contents.

        Test that the tarball has exactly expected contents. The basedir
        is checked for and prepended if it is not None. The mode is the mode
        used in tarfile.open defaults to r:gz. If skip_basedir is True and
        basedir is not None then the basedir wont be tested for itself.
        """
        if basedir is None:
            real_expected = expected[:]
        else:
            if skip_basedir:
                real_expected = []
            else:
                real_expected = [basedir]
        for item in expected:
            real_expected.append(os.path.join(basedir, item).rstrip("/"))
        extras = []
        tar = tarfile.open(tarball, 'r:gz')
        try:
            for tarinfo in tar:
                if tarinfo.name in real_expected:
                    index = real_expected.index(tarinfo.name)
                    del real_expected[index:index+1]
                else:
                    extras.append(tarinfo.name)

            if len(real_expected) > 0:
                self.fail("Files not found in %s: %s"
                          % (tarball, ", ".join(real_expected)))
            if len(extras) > 0:
                self.fail("Files not expected to be found in %s: %s"
                          % (tarball, ", ".join(extras)))
        finally:
            tar.close()


class SourcePackageBuilder(object):
    """An interface to ease building source packages.

    >>> builder = SourcePackageBuilder("package", Version("0.1-1"))
    >>> builder.add_upstream_file("foo")
    >>> builder.add_debian_file("debian/copyright")
    >>> builder.add_default_control()
    >>> builder.build()
    >>> builder.new_version(Version("0.2-1"))
    >>> builder.add_upstream_file("bar")
    >>> builder.remove_upstream_file("foo")
    >>> builder.build()
    >>> builder.dsc_name()
    """

    def __init__(self, name, version, native=False, version3=False,
                 multiple_upstream_tarballs=None):
        """
        :param name: Package name
        :param version: Package version
        :param native: Whether to build a native source package
        :param version3: Whether to build a version 3.0 source package
        :param multiple_upstream_tarballs: A list of each top-level directory
            within the upstream tree which is to be packed as a source format
            3.0 (quilt) additional upstream tarball
        """
        self.upstream_files = {}
        self.upstream_symlinks = {}
        self.debian_files = {}
        self.name = name
        self.native = native
        self.version3 = version3
        self.multiple_upstream_tarballs = multiple_upstream_tarballs
        if multiple_upstream_tarballs and not (version3 and not native):
            raise AssertionError(
                "Multiple upstream tarballs are only "
                "possible with 3.0 (quilt) format")
        self._cl = Changelog()
        self.new_version(version)

    def add_upstream_file(self, name, content=None):
        self.add_upstream_files([(name, content)])

    def add_upstream_files(self, files):
        for new_file in files:
            self.upstream_files[new_file[0]] = new_file[1]

    def add_upstream_symlink(self, name, target):
        self.upstream_symlinks[name] = target

    def remove_upstream_file(self, filename):
        del self.upstream_files[filename]

    def add_debian_file(self, name, content=None):
        self.add_debian_files([(name, content)])

    def add_debian_files(self, files):
        for new_file in files:
            self.debian_files[new_file[0]] = new_file[1]

    def remove_debian_file(self, filename):
        del self.debian_files[filename]

    def add_default_control(self):
        text = """Source: %s\nSection: misc\n""" % self.name
        text += "Priority: optional\n"
        text += "Maintainer: Maintainer <nobody@ubuntu.com>\n"
        text += "\n"
        text += "Package: %s\n" % self.name
        text += "Architecture: all\n\n"
        self.add_debian_file("debian/control", text)

    def new_version(self, version, change_text=None):
        self._cl.new_block(
                package=self.name, version=version,
                distributions="unstable", urgency="low",
                author="Maint <maint@maint.org>",
                date="Wed, 19 Mar 2008 21:27:37 +0000")
        if change_text is None:
            self._cl.add_change("  * foo")
        else:
            self._cl.add_change(change_text)

    def dsc_name(self):
        return "%s_%s.dsc" % (self.name, str(self._cl.version))

    def tar_name(self):
        if self.native:
            return "%s_%s.tar.gz" % (self.name, str(self._cl.version))
        return "%s_%s.orig.tar.gz" % (
            self.name, str(self._cl.version.upstream_version))

    def diff_name(self):
        assert not self.native, "Can't have a diff with a native package"
        return "%s_%s.diff.gz" % (self.name, str(self._cl.version))

    def changes_name(self):
        return "%s_%s_source.changes" % (self.name, str(self._cl.version))

    def _make_files(self, files_list, basedir):
        for (path, content) in files_list.items():
            dirname = os.path.dirname(path)
            if dirname is not None and dirname != "":
                if not os.path.exists(os.path.join(basedir, dirname)):
                    os.makedirs(os.path.join(basedir, dirname))
            with open(os.path.join(basedir, path), 'w') as f:
                if content is None:
                    content = ''
                f.write(content)

    def _make_symlinks(self, files_list, basedir):
        for (path, target) in files_list.items():
            dirname = os.path.dirname(path)
            if dirname is not None and dirname != "":
                if not os.path.exists(os.path.join(basedir, dirname)):
                    os.makedirs(os.path.join(basedir, dirname))
            os.symlink(target, os.path.join(basedir, path))

    def basedir(self):
        return self.name + "-" + str(self._cl.version.upstream_version)

    def write_debian_files(self, basedir):
        self._make_files(self.debian_files, basedir)
        self._make_files({"debian/changelog": str(self._cl)}, basedir)

    def _make_base(self):
        basedir = self.basedir()
        os.mkdir(basedir)
        self._make_files(self.upstream_files, basedir)
        self._make_symlinks(self.upstream_symlinks, basedir)
        return basedir

    def build(self, tar_format=None):
        if tar_format is None:
            tar_format = 'gz'
        basedir = self._make_base()
        if not self.version3:
            if not self.native:
                orig_basedir = basedir + ".orig"
                shutil.copytree(basedir, orig_basedir, symlinks=True)
                cmd = ["dpkg-source", "-sa", "-b", basedir]
                if os.path.exists(
                    "%s_%s.orig.tar.gz" % (
                        self.name, self._cl.version.upstream_version)):
                    cmd = ["dpkg-source", "-ss", "-b", basedir]
            else:
                cmd = ["dpkg-source", "-sn", "-b", basedir]
        else:
            if not self.native:
                if self.multiple_upstream_tarballs:
                    for part in self.multiple_upstream_tarballs:
                        tar_path = "%s_%s.orig-%s.tar.%s" % (
                            self.name, self._cl.version.upstream_version,
                            part, tar_format)
                        if os.path.exists(tar_path):
                            os.unlink(tar_path)
                        tar = tarfile.open(tar_path, 'w:%s' % tar_format)
                        part_basedir = os.path.join(basedir, part)
                        try:
                            tar.add(part_basedir, arcname=part)
                        finally:
                            tar.close()
                        shutil.rmtree(part_basedir)
                tar_path = "%s_%s.orig.tar.%s" % (
                    self.name, self._cl.version.upstream_version, tar_format)
                if os.path.exists(tar_path):
                    os.unlink(tar_path)
                tar = tarfile.open(tar_path, 'w:%s' % tar_format)
                try:
                    tar.add(basedir)
                finally:
                    tar.close()
                cmd = ["dpkg-source", "--format=3.0 (quilt)", "-b",
                       basedir]
            else:
                cmd = ["dpkg-source", "--format=3.0 (native)", "-b",
                       basedir]
        self.write_debian_files(basedir)
        proc = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        ret = proc.wait()
        assert ret == 0, "dpkg-source failed, output:\n%s" % \
            (proc.stdout.read(),)
        cmd = "dpkg-genchanges -S > ../%s" % self.changes_name()
        proc = subprocess.Popen(
            cmd, shell=True, stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT, cwd=basedir)
        ret = proc.wait()
        assert ret == 0, "dpkg-genchanges failed, output:\n%s" % \
            (proc.stdout.read(),)
        shutil.rmtree(basedir)


XzFeature = ExecutableFeature("xz")
LzmaFeature = ModuleAvailableFeature("lzma")



#    util.py -- Utility functions
#    Copyright (C) 2008 Canonical Ltd.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from __future__ import absolute_import

try:
    from ...revisionspec import InvalidRevisionSpec
except ImportError:  # breezy < 3.2
    from ...errors import InvalidRevisionSpec

from ...errors import (
    NoSuchTag,
    BzrError,
    )
from ...revisionspec import RevisionSpec, RevisionInfo


class UnknownVersion(BzrError):
    _fmt = ('No tag exists in this branch indicating that version '
            '"%(version)s" has been uploaded.')

    def __init__(self, version):
        BzrError.__init__(self, version=version)


class VersionNotSpecified(BzrError):
    _fmt = "You did not specify a package version."


class RevisionSpec_package(RevisionSpec):
    """Selects a revision based on the version of the package."""

    help_txt = """Selects the revision corresponding to a version of a package.

    Given a package version number this revision specifier will allow you
    specify the revision corresponding to the upload of that version of
    the package.
    """
    wants_revision_history = False
    prefix = 'package:'

    def _match_on(self, branch, revs=None):
        version_spec = self.spec
        dist_spec = None  # noqa: F841

        if version_spec == '':
            raise VersionNotSpecified

        try:
            revision_id = branch.tags.lookup_tag(version_spec)
            return RevisionInfo.from_revision_id(
                branch, revision_id)
        except NoSuchTag:
            raise UnknownVersion(version_spec)


class RevisionSpec_upstream(RevisionSpec):
    """Selects the matching upstream revision."""

    help_txt = """Selects the revision matching the upstream specified
    or the current upstream.

    This will look at debian/changelog to find out the current version,
    and then look up the upstream.
    """
    wants_revision_history = False
    prefix = 'upstream:'

    def _match_on(self, branch, revs=None):
        from ...workingtree import WorkingTree
        from .util import (
            find_changelog,
            MissingChangelogError,
            )
        from .upstream.pristinetar import get_pristine_tar_source
        from .upstream import StackedUpstreamSource
        from debian.changelog import Version
        tree, subpath = WorkingTree.open_containing('.')
        try:
            (cl, top_level) = find_changelog(tree, subpath, merge=False)
        except MissingChangelogError as e:
            raise InvalidRevisionSpec(
                self.user_spec, branch,
                "no debian/changelog file found: %s" % e)
        if self.spec == '':
            version_spec = cl.version.upstream_version
            if not cl.version.debian_version:
                raise InvalidRevisionSpec(
                    self.user_spec, branch,
                    "This is a native package.")
        else:
            version = Version(self.spec)
            if version.upstream_version:
                version_spec = version.upstream_version
            else:
                version_spec = self.spec

        upstream_source = StackedUpstreamSource([
            get_pristine_tar_source(tree, branch),
            ])
        try:
            revision_id = upstream_source.version_as_revisions(
                cl.package, version_spec)[None]
            return RevisionInfo.from_revision_id(
                branch, revision_id)
        except NoSuchTag:
            raise UnknownVersion(version_spec)



#    test_dep3.py -- Testsuite for builddeb dep3.py
#    Copyright (C) 2011 Canonical Ltd.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from __future__ import absolute_import

from email.message import Message
from email.parser import Parser
from io import BytesIO, TextIOWrapper

from ....revision import (
    NULL_REVISION,
    )
from io import StringIO

from ....tests import (
    TestCase,
    )

from ..dep3 import (
    describe_origin,
    determine_applied_upstream,
    gather_bugs_and_authors,
    write_dep3_bug_line,
    write_dep3_patch,
    write_dep3_patch_header,
    )

from . import (
    TestCaseWithTransport,
    )


class Dep3HeaderTests(TestCase):

    def dep3_header(
            self, description=None, origin=None, forwarded=None,
            bugs=None, authors=None, revision_id=None, last_update=None,
            applied_upstream=None):
        f = StringIO()
        write_dep3_patch_header(
            f, description=description, origin=origin,
            forwarded=forwarded, bugs=bugs, authors=authors,
            revision_id=revision_id, last_update=last_update,
            applied_upstream=applied_upstream)
        f.seek(0)
        return Parser().parse(f)

    def test_description(self):
        ret = self.dep3_header(description="This patch fixes the foobar")
        self.assertEquals("This patch fixes the foobar", ret["Description"])

    def test_last_updated(self):
        ret = self.dep3_header(last_update=1304840034)
        self.assertEquals("2011-05-08", ret["Last-Update"])

    def test_revision_id(self):
        ret = self.dep3_header(revision_id=b"myrevid")
        self.assertEquals("myrevid", ret["X-Bzr-Revision-Id"])

    def test_authors(self):
        authors = [
            "Jelmer Vernooij <jelmer@canonical.com>",
            "James Westby <james.westby@canonical.com>"]
        ret = self.dep3_header(authors=authors)
        self.assertEquals([
            "Jelmer Vernooij <jelmer@canonical.com>",
            "James Westby <james.westby@canonical.com>"],
            ret.get_all("Author"))

    def test_origin(self):
        ret = self.dep3_header(origin="Cherrypick from upstream")
        self.assertEquals("Cherrypick from upstream", ret["Origin"])

    def test_forwarded(self):
        ret = self.dep3_header(forwarded="not needed")
        self.assertEquals("not needed", ret["Forwarded"])

    def test_applied_upstream(self):
        ret = self.dep3_header(applied_upstream="commit 45")
        self.assertEquals("commit 45", ret["Applied-Upstream"])

    def test_bugs(self):
        bugs = [
            ("http://bugs.debian.org/424242", "fixed"),
            ("https://bugs.launchpad.net/bugs/20110508", "fixed"),
            ("http://bugzilla.samba.org/bug.cgi?id=52", "fixed")]
        ret = self.dep3_header(bugs=bugs)
        self.assertEquals([
            "https://bugs.launchpad.net/bugs/20110508",
            "http://bugzilla.samba.org/bug.cgi?id=52"],
            ret.get_all("Bug"))
        self.assertEquals(
            ["http://bugs.debian.org/424242"],
            ret.get_all("Bug-Debian"))

    def test_write_bug_fix_only(self):
        # non-fixed bug lines are ignored
        message = Message()
        write_dep3_bug_line(message, "http://bar/", "pending")
        self.assertEquals("\n", str(message))

    def test_write_normal_bug(self):
        message = Message()
        write_dep3_bug_line(
            message, "http://bugzilla.samba.org/bug.cgi?id=42", "fixed")
        self.assertEquals(
            "Bug: http://bugzilla.samba.org/bug.cgi?id=42\n\n",
            str(message))

    def test_write_debian_bug(self):
        message = Message()
        write_dep3_bug_line(message, "http://bugs.debian.org/234354", "fixed")
        self.assertEquals(
            "Bug-Debian: http://bugs.debian.org/234354\n\n",
            str(message))


class GatherBugsAndAuthors(TestCaseWithTransport):

    def test_none(self):
        branch = self.make_branch(".")
        self.assertEquals(
            (set(), set(), None),
            gather_bugs_and_authors(branch.repository, []))

    def test_multiple_authors(self):
        tree = self.make_branch_and_tree(".")
        revid1 = tree.commit(
            authors=["Jelmer Vernooij <jelmer@canonical.com>"],
            timestamp=1304844311, message="msg")
        revid2 = tree.commit(
            authors=["Max Bowsher <maxb@f2s.com>"],
            timestamp=1304844278, message="msg")
        self.assertEquals((set(), set([
            "Jelmer Vernooij <jelmer@canonical.com>",
            "Max Bowsher <maxb@f2s.com>"]), 1304844311),
            gather_bugs_and_authors(tree.branch.repository, [revid1, revid2]))

    def test_bugs(self):
        tree = self.make_branch_and_tree(".")
        revid1 = tree.commit(
            authors=["Jelmer Vernooij <jelmer@canonical.com>"],
            timestamp=1304844311, message="msg",
            revprops={"bugs": "http://bugs.samba.org/bug.cgi?id=2011 fixed\n"})
        self.assertEquals((
            set([("http://bugs.samba.org/bug.cgi?id=2011", "fixed")]),
            set(["Jelmer Vernooij <jelmer@canonical.com>"]), 1304844311),
            gather_bugs_and_authors(tree.branch.repository, [revid1]))


class DetermineAppliedUpstreamTests(TestCaseWithTransport):

    def test_not_applied(self):
        upstream = self.make_branch_and_tree("upstream")
        feature = self.make_branch_and_tree("feature")
        feature.commit(message="every bloody emperor")
        self.addCleanup(feature.lock_read().unlock)
        self.assertEquals(
            "no", determine_applied_upstream(upstream.branch, feature.branch))

    def test_merged(self):
        upstream = self.make_branch_and_tree("upstream")
        upstream.commit(message="initial upstream commit")
        feature = upstream.controldir.sprout("feature").open_workingtree()
        feature.commit(message="nutter alert")
        upstream.merge_from_branch(feature.branch)
        upstream.commit(message="merge feature")
        self.addCleanup(upstream.lock_read().unlock)
        self.addCleanup(feature.lock_read().unlock)
        self.assertEquals(
            "merged in revision 2",
            determine_applied_upstream(upstream.branch, feature.branch))


class DescribeOriginTests(TestCaseWithTransport):

    def test_no_public_branch(self):
        tree = self.make_branch_and_tree(".")
        revid1 = tree.commit(message="msg1")
        self.assertEquals(
            "commit, revision id: %s" % revid1.decode('utf-8'),
            describe_origin(tree.branch, revid1))

    def test_public_branch(self):
        tree = self.make_branch_and_tree(".")
        tree.branch.set_public_branch("http://example.com/public")
        revid1 = tree.commit(message="msg1")
        self.assertEquals(
            "commit, http://example.com/public, revision: 1",
            describe_origin(tree.branch, revid1))


class FullDep3PatchTests(TestCaseWithTransport):

    def test_simple(self):
        f = StringIO()
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("foo", "data")])
        tree.add("foo")
        revid = tree.commit(
            "msg", rev_id=b"arevid", timestamp=1304849661, timezone=0)
        write_dep3_patch(
            f, tree.branch, NULL_REVISION, revid,
            description="Nutter alert",
            forwarded="not needed",
            authors=set(["Jelmer <jelmer@samba.org>"]))
        self.assertEqual(
            "Description: Nutter alert\n"
            "Forwarded: not needed\n"
            "Author: Jelmer <jelmer@samba.org>\n"
            "X-Bzr-Revision-Id: arevid\n"
            "\n"
            "=== added file 'foo'\n"
            "--- old/foo\t1970-01-01 00:00:00 +0000\n"
            "+++ new/foo\t2011-05-08 10:14:21 +0000\n"
            "@@ -0,0 +1,1 @@\n"
            "+data\n"
            "\\ No newline at end of file\n"
            "\n", f.getvalue())



#    util.py -- Utility functions
#    Copyright (C) 2006 James Westby <jw+debian@jameswestby.net>
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from __future__ import absolute_import

import errno
import hashlib
import signal
import shutil
import subprocess
import tempfile
import os
import re

from debian import deb822
from debian.changelog import Changelog, ChangelogParseError
from debian.copyright import Copyright, NotMachineReadableError

from debmutate.changelog import (
    changes_by_author,
    find_thanks,
    find_extra_authors,
    find_last_distribution,
    strip_changelog_message,
    )

from ... import (
    bugtracker,
    errors,
    osutils,
    urlutils,
    )
from ...trace import (
    mutter,
    note,
    warning,
    )
from ...transport import (
    do_catching_redirections,
    get_transport,
    )
from . import (
    global_conf,
    )
from .config import (
    DebBuildConfig,
    BUILD_TYPE_MERGE,
    BUILD_TYPE_NATIVE,
    BUILD_TYPE_NORMAL,
    )
from .errors import (
    BzrError,
    )

BUILDDEB_DIR = '.bzr-builddeb'

NEW_LOCAL_CONF = 'debian/local.conf.local'
NEW_CONF = 'debian/bzr-builddeb.conf'
DEFAULT_CONF = os.path.join(BUILDDEB_DIR, 'default.conf')
LOCAL_CONF = os.path.join(BUILDDEB_DIR, 'local.conf')


class MissingChangelogError(BzrError):
    _fmt = 'Could not find changelog at %(location)s in tree.'

    def __init__(self, locations):
        BzrError.__init__(self, location=locations)


_DEBIAN_RELEASES = None
_UBUNTU_RELEASES = None


def _get_release_names():
    global _DEBIAN_RELEASES, _UBUNTU_RELEASES
    try:
        from distro_info import DebianDistroInfo, UbuntuDistroInfo
    except ImportError:
        warning(
            "distro_info not available. Unable to retrieve current "
            "list of releases.")
        _DEBIAN_RELEASES = []
        _UBUNTU_RELEASES = []
    else:
        # distro info is not available
        _DEBIAN_RELEASES = DebianDistroInfo().all
        _UBUNTU_RELEASES = UbuntuDistroInfo().all

    _DEBIAN_RELEASES.extend(['stable', 'testing', 'unstable', 'frozen'])


def debian_releases():
    if _DEBIAN_RELEASES is None:
        _get_release_names()
    return _DEBIAN_RELEASES


def ubuntu_releases():
    if _UBUNTU_RELEASES is None:
        _get_release_names()
    return _UBUNTU_RELEASES


DEBIAN_POCKETS = ('', '-security', '-proposed-updates', '-backports')
UBUNTU_POCKETS = ('', '-proposed', '-updates', '-security', '-backports')


def safe_decode(s):
    """Decode a string into a Unicode value."""
    if isinstance(s, str):  # Already unicode
        mutter('safe_decode() called on an already-decoded string: %r' % (s,))
        return s
    try:
        return s.decode('utf-8')
    except UnicodeDecodeError as e:
        mutter('safe_decode(%r) falling back to iso-8859-1: %s' % (s, e))
        # TODO: Looking at BeautifulSoup it seems to use 'chardet' to try to
        #       guess the encoding of a given text stream. We might want to
        #       take a closer look at that.
        # TODO: Another possibility would be to make the fallback encoding
        #       configurable, possibly exposed as a command-line flag, for now,
        #       this seems 'good enough'.
        return s.decode('iso-8859-1')


def recursive_copy(fromdir, todir):
    """Copy the contents of fromdir to todir.

    Like shutil.copytree, but the destination directory must already exist
    with this method, rather than not exists for shutil.
    """
    mutter("Copying %s to %s", fromdir, todir)
    for entry in os.listdir(fromdir):
        path = os.path.join(fromdir, entry)
        if os.path.isdir(path):
            tosubdir = os.path.join(todir, entry)
            if not os.path.exists(tosubdir):
                os.mkdir(tosubdir)
            recursive_copy(path, tosubdir)
        else:
            # Python 3 has a follow_symlinks argument to shutil.copy, but
            # Python 2 does not...
            if os.path.islink(path):
                os.symlink(os.readlink(path), os.path.join(todir, entry))
            else:
                shutil.copy(path, todir)


class AddChangelogError(BzrError):
    _fmt = 'Please add "%(changelog)s" to the branch using bzr add.'

    def __init__(self, changelog):
        BzrError.__init__(self, changelog=changelog)


def find_changelog(t, subpath='', merge=False, max_blocks=1):
    """Find the changelog in the given tree.

    First looks for 'debian/changelog'. If "merge" is true will also
    look for 'changelog'.

    The returned changelog is created with 'allow_empty_author=True'
    as some people do this but still want to build.
    'max_blocks' defaults to 1 to try and prevent old broken
    changelog entries from causing the command to fail.

    "top_level" is a subset of "merge" mode. It indicates that the
    '.bzr' dir is at the same level as 'changelog' etc., rather
    than being at the same level as 'debian/'.

    :param t: the Tree to look in.
    :param merge: whether this is a "merge" package.
    :param max_blocks: Number of max_blocks to parse (defaults to 1). Use None
        to parse the entire changelog.
    :return: (changelog, top_level) where changelog is the Changelog,
        and top_level is a boolean indicating whether the file is
        located at 'changelog' (rather than 'debian/changelog') if
        merge was given, False otherwise.
    """
    top_level = False
    with t.lock_read():
        changelog_file = osutils.pathjoin(subpath, 'debian/changelog')
        if not t.has_filename(changelog_file):
            checked_files = [changelog_file]
            if merge:
                # Assume LarstiQ's layout (.bzr in debian/)
                changelog_file = osutils.pathjoin(subpath, 'changelog')
                top_level = True
                if not t.has_filename(changelog_file):
                    checked_files.append(changelog_file)
                    changelog_file = None
            else:
                changelog_file = None
            if changelog_file is None:
                if getattr(t, "abspath", None):
                    checked_files = [t.abspath(f) for f in checked_files]
                raise MissingChangelogError(" or ".join(checked_files))
        elif merge and t.has_filename(osutils.pathjoin(subpath, 'changelog')):
            # If it is a "top_level" package and debian is a symlink to
            # "." then it will have found debian/changelog. Try and detect
            # this.
            debian_file = osutils.pathjoin(subpath, 'debian')
            if (t.is_versioned(debian_file) and
                    t.kind(debian_file) == 'symlink' and
                    t.get_symlink_target(debian_file) == '.'):
                changelog_file = 'changelog'
                top_level = True
        mutter("Using '%s' to get package information", changelog_file)
        if not t.is_versioned(changelog_file):
            raise AddChangelogError(changelog_file)
        contents = t.get_file_text(changelog_file)
    changelog = Changelog()
    changelog.parse_changelog(
        contents, max_blocks=max_blocks, allow_empty_author=True)
    return changelog, top_level


def tarball_name(package, version, component=None, format=None):
    """Return the name of the .orig.tar.gz for the given package and version.

    :param package: the name of the source package.
    :param version: the upstream version of the package.
    :param component: Component name (None for base)
    :param format: the format for the tarball. If None then 'gz' will be
         used. You probably want on of 'gz', 'bz2', 'lzma' or 'xz'.
    :return: a string that is the name of the upstream tarball to use.
    """
    if format is None:
        format = 'gz'
    name = "%s_%s.orig" % (package, str(version))
    if component is not None:
        name += "-" + component
    return "%s.tar.%s" % (name, format)


def suite_to_distribution(suite):
    """Infer the distribution from a suite.

    When passed the name of a suite (anything in the distributions field of
    a changelog) it will infer the distribution from that (i.e. Debian or
    Ubuntu).

    :param suite: the string containing the suite
    :return: "debian", "ubuntu", or None if the distribution couldn't be
        inferred.
    """
    all_debian = [r + t for r in debian_releases() for t in DEBIAN_POCKETS]
    all_ubuntu = [r + t for r in ubuntu_releases() for t in UBUNTU_POCKETS]
    if suite in all_debian:
        return "debian"
    if suite in all_ubuntu:
        return "ubuntu"
    if suite == 'kali' or suite.startswith('kali-'):
        return "kali"
    return None


def lookup_distribution(distribution_or_suite):
    """Get the distribution name based on a distribution or suite name.

    :param distribution_or_suite: a string that is either the name of
        a distribution or a suite.
    :return: a string with a distribution name or None.
    """
    if distribution_or_suite.lower() in ("debian", "ubuntu"):
        return distribution_or_suite.lower()
    return suite_to_distribution(distribution_or_suite)


def md5sum_filename(filename):
    """Calculate the md5sum of a file by name.

    :param filename: Path of the file to checksum
    :return: MD5 Checksum as hex digest
    """
    m = hashlib.md5()
    with open(filename, 'rb') as f:
        for line in f:
            m.update(line)
    return m.hexdigest()


def move_file_if_different(source, target, md5sum):
    """Overwrite a file if its new contents would be different from the current
    contents.

    :param source: Path of the source file
    :param target: Path of the target file
    :param md5sum: MD5Sum (as hex digest) of the source file
    """
    if os.path.exists(target):
        if os.path.samefile(source, target):
            return
        t_md5sum = md5sum_filename(target)
        if t_md5sum == md5sum:
            return
    shutil.move(source, target)


def write_if_different(contents, target):
    """(Over)write a file with `contents` if they are different from its
    current content.

    :param contents: The contents to write, as a string
    :param target: Path of the target file
    """
    md5sum = hashlib.md5()
    md5sum.update(contents)
    fd, temp_path = tempfile.mkstemp("builddeb-rename-")
    fobj = os.fdopen(fd, "wb")
    try:
        try:
            fobj.write(contents)
        finally:
            fobj.close()
        move_file_if_different(temp_path, target, md5sum.hexdigest())
    finally:
        if os.path.exists(temp_path):
            os.unlink(temp_path)


def _download_part(name, base_transport, target_dir, md5sum):
    part_base_dir, part_path = urlutils.split(name)
    f_t = base_transport
    if part_base_dir != '':
        f_t = base_transport.clone(part_base_dir)
    with f_t.get(part_path) as f_f:
        target_path = os.path.join(target_dir, part_path)
        fd, temp_path = tempfile.mkstemp(prefix="builddeb-")
        fobj = os.fdopen(fd, "wb")
        try:
            try:
                shutil.copyfileobj(f_f, fobj)
            finally:
                fobj.close()
            move_file_if_different(temp_path, target_path, md5sum)
        finally:
            if os.path.exists(temp_path):
                os.unlink(temp_path)


def open_file(url):
    """Open a file from a URL.

    :param url: URL to open
    :return: A file-like object.
    """
    filename, transport = open_transport(url)
    return open_file_via_transport(filename, transport)


def open_transport(path):
    """Obtain an appropriate transport instance for the given path."""
    base_dir, path = urlutils.split(path)
    transport = get_transport(base_dir)
    return (path, transport)


def open_file_via_transport(filename, transport):
    """Open a file using the transport, follow redirects as necessary."""
    def open_file(transport):
        return transport.get(filename)

    def follow_redirection(transport, e, redirection_notice):
        mutter(redirection_notice)
        _filename, redirected_transport = open_transport(e.target)
        return redirected_transport

    result = do_catching_redirections(open_file, transport, follow_redirection)
    return result


def _dget(cls, dsc_location, target_dir):
    """Copy all files referenced by a .dsc file.

    Args:
      cls: Parser class
      dsc_location: Source file location
      target_dir: Target directory
    Return:
      path to target source file
    """
    if not os.path.isdir(target_dir):
        raise errors.NotADirectory(target_dir)
    path, dsc_t = open_transport(dsc_location)
    with open_file_via_transport(path, dsc_t) as f:
        dsc_contents = f.read()
    dsc = cls(dsc_contents)
    for file_details in dsc['files']:
        name = file_details['name']
        _download_part(name, dsc_t, target_dir, file_details['md5sum'])
    target_file = os.path.join(target_dir, path)
    write_if_different(dsc_contents, target_file)
    return target_file


def dget(dsc_location, target_dir):
    return _dget(deb822.Dsc, dsc_location, target_dir)


def dget_changes(changes_location, target_dir):
    return _dget(deb822.Changes, changes_location, target_dir)


def get_parent_dir(target):
    parent = os.path.dirname(target)
    if os.path.basename(target) == '':
        parent = os.path.dirname(parent)
    return parent


def find_bugs_fixed(changes, branch, _lplib=None):
    """Find the bugs marked fixed in a changelog entry.

    :param changes: A list of the contents of the changelog entry.
    :param branch: Bazaar branch associated with the package
    :return: String with bugs closed, as appropriate for a Bazaar "bugs"
        revision property.
    """
    if _lplib is None:
        from . import launchpad as _lplib
    bugs = []
    for new_author, linenos, lines in changes_by_author(changes):
        for match in re.finditer(
                "closes:\\s*(?:bug)?\\#?\\s?\\d+"
                "(?:,\\s*(?:bug)?\\#?\\s?\\d+)*", ''.join(lines),
                re.IGNORECASE):
            closes_list = match.group(0)
            for match in re.finditer("\\d+", closes_list):
                bug_url = bugtracker.get_bug_url("deb", branch, match.group(0))
                bugs.append(bug_url + " fixed")
                lp_bugs = _lplib.ubuntu_bugs_for_debian_bug(match.group(0))
                if len(lp_bugs) == 1:
                    bug_url = bugtracker.get_bug_url("lp", branch, lp_bugs[0])
                    bugs.append(bug_url + " fixed")
        for match in re.finditer(
                "lp:\\s+\\#\\d+(?:,\\s*\\#\\d+)*",
                ''.join(lines), re.IGNORECASE):
            closes_list = match.group(0)
            for match in re.finditer("\\d+", closes_list):
                bug_url = bugtracker.get_bug_url("lp", branch, match.group(0))
                bugs.append(bug_url + " fixed")
                deb_bugs = _lplib.debian_bugs_for_ubuntu_bug(match.group(0))
                if len(deb_bugs) == 1:
                    bug_url = bugtracker.get_bug_url(
                        "deb", branch, deb_bugs[0])
                    bugs.append(bug_url + " fixed")
    return bugs


def get_commit_info_from_changelog(changelog, branch, _lplib=None):
    """Retrieves the messages from the last section of debian/changelog.

    Reads the latest stanza of debian/changelog and returns the
    text of the changes in that section. It also returns other
    information about the change, including the authors of the change,
    anyone that is thanked, and the bugs that are declared fixed by it.

    :return: a tuple (message, authors, thanks, bugs). message is the
        commit message that should be used. authors is a list of strings,
        with those that contributed to the change, thanks is a list
        of string, with those who were thanked in the changelog entry.
        bugs is a list of bug URLs like for --fixes.
        If the information is not available then any can be None.
    """
    message = None
    authors = []
    thanks = []
    bugs = []
    if changelog._blocks:
        block = changelog._blocks[0]
        authors = [safe_decode(block.author)]
        authors += find_extra_authors(block.changes())
        changes = strip_changelog_message(block.changes())
        bugs = find_bugs_fixed(block.changes(), branch, _lplib=_lplib)
        thanks = find_thanks(block.changes())
        message = safe_decode("\n".join(changes).replace("\r", ""))
    return (message, authors, thanks, bugs)


def subprocess_setup():
    # Python installs a SIGPIPE handler by default. This is usually not what
    # non-Python subprocesses expect.
    # Many, many thanks to Colin Watson
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)


def debuild_config(tree, subpath):
    """Obtain the Debuild configuration object.

    :param tree: A Tree object, can be a WorkingTree or RevisionTree.
    """
    config_files = []
    user_config = None
    if subpath in ('.', None):
        subpath = ''
    new_local_conf = osutils.pathjoin(subpath, NEW_LOCAL_CONF)
    local_conf = osutils.pathjoin(subpath, LOCAL_CONF)
    new_conf = osutils.pathjoin(subpath, NEW_CONF)
    default_conf = osutils.pathjoin(subpath, DEFAULT_CONF)

    if tree.has_filename(new_local_conf):
        if not tree.is_versioned(new_local_conf):
            config_files.append(
                (tree.get_file(new_local_conf), True, "local.conf"))
        else:
            warning('Not using configuration from %s as it is versioned.',
                    new_local_conf)
    if tree.has_filename(local_conf):
        if not tree.is_versioned(local_conf):
            config_files.append(
                (tree.get_file(local_conf), True, "local.conf"))
        else:
            warning('Not using configuration from %s as it is versioned.',
                    local_conf)
    config_files.append((global_conf(), True))
    user_config = global_conf()
    if tree.is_versioned(new_conf):
        config_files.append(
            (tree.get_file(new_conf), False, "bzr-builddeb.conf"))
    if tree.is_versioned(default_conf):
        config_files.append(
            (tree.get_file(default_conf), False, "default.conf"))
    config = DebBuildConfig(config_files, tree=tree)
    config.set_user_config(user_config)
    return config


class UnableToFindPreviousUpload(BzrError):

    _fmt = ("Unable to determine the previous upload for --package-merge.")


def find_previous_upload(tree, subpath, merge=False):
    """Given a tree, find the previous upload to the distribution.

    When e.g. Ubuntu merges from Debian they want to build with
    -vPREV_VERSION. Here's where we find that previous version.

    We look at the last changelog entry and find the upload target.
    We then search backwards until we find the same target. That's
    the previous version that we return.

    We require there to be a previous version, otherwise we throw
    an error.

    It's not a simple string comparison to find the same target in
    a previous version, as we should consider old series in e.g.
    Ubuntu.
    """
    try:
        cl, top_level = find_changelog(tree, subpath, merge, max_blocks=None)
    except ChangelogParseError:
        raise UnableToFindPreviousUpload()
    return changelog_find_previous_upload(cl)


class NoPreviousUpload(BzrError):

    _fmt = ("There was no previous upload to %(distribution)s.")

    def __init__(self, distribution):
        BzrError.__init__(self, distribution=distribution)


def changelog_find_previous_upload(cl):
    """Find the version of the previous upload.

    :param cl: Changelog object
    :return: Version object for the previous upload
    :raise NoPreviousUpload: Raised when there is no previous upload
    """
    current_target = find_last_distribution(cl)
    all_debian = [r + t for r in debian_releases() for t in DEBIAN_POCKETS]
    all_ubuntu = [r + t for r in ubuntu_releases() for t in UBUNTU_POCKETS]
    if current_target in all_debian:
        match_targets = (current_target,)
    elif current_target in all_ubuntu:
        match_targets = ubuntu_releases()
        if "-" in current_target:
            match_targets += tuple(
                [current_target.split("-", 1)[0] + t for t in UBUNTU_POCKETS])
    else:
        # If we do not recognize the current target in order to apply special
        # rules to it, then just assume that only previous uploads to exactly
        # the same target count.
        match_targets = (current_target,)
    for block in cl._blocks[1:]:
        if block.distributions.split(" ")[0] in match_targets:
            return block.version
    raise NoPreviousUpload(current_target)


def tree_contains_upstream_source(tree, subpath=''):
    """Guess if the specified tree contains the upstream source.

    :param tree: A RevisionTree.
    :return: Boolean indicating whether or not the tree contains the upstream
        source. None if the tree is empty
    """
    present_files = set(
        [f[0] for f in tree.list_files(recursive=False, from_dir=subpath)
         if f[1] == 'V'])
    if len(present_files) == 0:
        return None
    packaging_files = frozenset([
        "debian", ".bzr-builddeb", ".bzrignore", ".gitignore"])
    return (len(present_files - packaging_files) > 0)


def tree_get_source_format(tree, subpath=''):
    """Retrieve the source format name from a package.

    :param path: Path to the package
    :return: String with package format
    """
    filename = osutils.pathjoin(subpath, "debian/source/format")
    try:
        text = tree.get_file_text(filename)
    except IOError as e:
        if e.errno == errno.ENOENT:
            return FORMAT_1_0
        raise
    except errors.NoSuchFile:
        return FORMAT_1_0
    return text.strip().decode('ascii')


FORMAT_1_0 = "1.0"
FORMAT_3_0_QUILT = "3.0 (quilt)"
FORMAT_3_0_NATIVE = "3.0 (native)"

NATIVE_SOURCE_FORMATS = [FORMAT_3_0_NATIVE]
NORMAL_SOURCE_FORMATS = [FORMAT_3_0_QUILT]


class InconsistentSourceFormatError(BzrError):

    _fmt = ("Inconsistency between source format and version: version is "
            "%(version_bool)snative, format is %(format_bool)snative.")

    def __init__(self, version_native, format_native):
        if version_native:
            version_bool = ""
        else:
            version_bool = "not "
        if format_native:
            format_bool = ""
        else:
            format_bool = "not "
        BzrError.__init__(
            self, version_bool=version_bool, format_bool=format_bool)


def guess_build_type(tree, version, subpath='', contains_upstream_source=True):
    """Guess the build type based on the contents of a tree.

    :param tree: A `Tree` object.
    :param version: `Version` of the upload.
    :param contains_upstream_source: Whether this branch contains the upstream
        source.
    :return: A build_type value.
    """
    source_format = tree_get_source_format(tree, subpath)
    if source_format in NATIVE_SOURCE_FORMATS:
        format_native = True
    elif source_format in NORMAL_SOURCE_FORMATS:
        format_native = False
    else:
        format_native = None

    if version is not None:
        version_native = (not version.debian_version)
    else:
        version_native = None

    # If the package doesn't have a debian revision then it is very probably
    # native, but it *could* be native.
    if type(version_native) is bool and type(format_native) is bool:
        if version_native is True and format_native is False:
            raise InconsistentSourceFormatError(version_native, format_native)
        if version_native is False and format_native is True:
            warning(
                'Version suggests non-native package, '
                'but format is for native.')

    if version_native or format_native:
        return BUILD_TYPE_NATIVE
    if contains_upstream_source is False:
        # Default to merge mode if there's only a debian/ directory
        return BUILD_TYPE_MERGE
    else:
        return BUILD_TYPE_NORMAL


def component_from_orig_tarball(tarball_filename, package, version):
    tarball_filename = os.path.basename(tarball_filename)
    prefix = "%s_%s.orig" % (package, version)
    if not tarball_filename.startswith(prefix):
        raise ValueError(
            "invalid orig tarball file %s does not have expected prefix %s" % (
                tarball_filename, prefix))
    base = tarball_filename[len(prefix):]
    for ext in (".tar.gz", ".tar.bz2", ".tar.lzma", ".tar.xz"):
        if tarball_filename.endswith(ext):
            base = base[:-len(ext)]
            break
    else:
        raise ValueError(
            "orig tarball file %s has unknown extension" % tarball_filename)
    if base == "":
        return None
    elif base[0] == "-":
        # Extra component
        return base[1:]
    else:
        raise ValueError(
                "Invalid extra characters in tarball filename %s" %
                tarball_filename)


class TarFailed(BzrError):
    _fmt = ("There was an error executing tar to %(operation)s %(tarball)s: "
            "%(error)s.")

    def __init__(self, operation, tarball, error):
        BzrError.__init__(
            self, operation=operation, tarball=tarball, error=error)


def needs_strip_components(tf):
    top_level_directories = set()
    for name in tf.getnames():
        top_level_directories.add(name.split('/')[0])
    return len(top_level_directories) == 1


def extract_orig_tarball(tarball_filename, component, target,
                         strip_components=None):
    """Extract an orig tarball.

    :param tarball: Path to the tarball
    :param component: Component name (or None for top-level)
    :param target: Target path
    """
    from tarfile import TarFile
    tar_args = ["tar"]
    if tarball_filename.endswith(".tar.bz2"):
        tar_args.append('xjf')
        tf = TarFile.bz2open(tarball_filename)
    elif (tarball_filename.endswith(".tar.lzma") or
          tarball_filename.endswith(".tar.xz")):
        tar_args.append('xJf')
        tf = TarFile.xzopen(tarball_filename)
    elif tarball_filename.endswith(".tar"):
        tar_args.append('xf')
        tf = TarFile.open(tarball_filename)
    elif (tarball_filename.endswith(".tar.gz") or
            tarball_filename.endswith(".tgz")):
        tf = TarFile.gzopen(tarball_filename)
        tar_args.append('xzf')
    else:
        note('Unable to figure out type of %s, '
             'assuming .tar.gz', tarball_filename)
        tf = TarFile.gzopen(tarball_filename)
        tar_args.append('xzf')

    try:
        if strip_components is None:
            if needs_strip_components(tf):
                strip_components = 1
            else:
                strip_components = 0
    finally:
        tf.close()
    if component is not None:
        target_path = os.path.join(target, component)
        os.mkdir(target_path)
    else:
        target_path = target
    tar_args.extend([tarball_filename, "-C", target_path])
    if strip_components is not None:
        tar_args.extend(["--strip-components", str(strip_components)])
    proc = subprocess.Popen(tar_args, preexec_fn=subprocess_setup,
                            stderr=subprocess.PIPE)
    (stdout, stderr) = proc.communicate()
    if proc.returncode != 0:
        raise TarFailed("extract", tarball_filename, error=stderr)


def extract_orig_tarballs(tarballs, target, strip_components=None):
    """Extract orig tarballs to a directory.

    :param tarballs: List of tarball filenames
    :param target: Target directory (must already exist)
    """
    for tarball_filename, component in tarballs:
        extract_orig_tarball(
            tarball_filename, component, target,
            strip_components=strip_components)


def dput_changes(path):
    """Upload a package."""
    (bd, changes_file) = os.path.split(path)
    subprocess.check_call(["dput", changes_file], cwd=bd)


def debsign(path, keyid=None):
    (bd, changes_file) = os.path.split(path)
    args = ["debsign"]
    if keyid:
        args.append("-k%s" % keyid)
    args.append(changes_file)
    subprocess.check_call(args, cwd=bd)


def changes_filename(package, version, arch):
    non_epoch_version = version.upstream_version
    if version.debian_version is not None:
        non_epoch_version += "-%s" % version.debian_version
    return "%s_%s_%s.changes" % (package, non_epoch_version, arch)


def get_build_architecture():
    try:
        return subprocess.check_output(
            ['dpkg-architecture', '-qDEB_BUILD_ARCH']).strip().decode()
    except subprocess.CalledProcessError as e:
        raise BzrError(
            "Could not find the build architecture: %s" % e)


def get_files_excluded(tree, subpath='', top_level=False):
    if top_level:
        path = os.path.join(subpath, 'copyright')
    else:
        path = os.path.join(subpath, 'debian', 'copyright')
    with tree.get_file(path) as f:
        try:
            copyright = Copyright(f, strict=False)
        except NotMachineReadableError:
            return []
        try:
            return copyright.header["Files-Excluded"].split()
        except KeyError:
            return []



#    directory.py -- Directory service that uses Debian Vcs-* fields
#    Copyright (C) 2008 Jelmer Vernooij <jelmer@samba.org>
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from __future__ import absolute_import

import re
import sys

from ... import urlutils
from ...directory_service import directories
from ...errors import DependencyNotPresent
from ...trace import note, warning

from debian.deb822 import Deb822
from debian.changelog import Version

from debmutate.vcs import source_package_vcs


def fixup_broken_git_url(url):
    """Attempt to fix up broken Git URLs.

    A common misspelling is to add an extra ":" after the hostname
    """
    (scheme, netloc, path, params,
     query, fragment) = urlutils.urlparse.urlparse(url, allow_fragments=False)
    if '@' in netloc:
        credentials, host = netloc.rsplit('@', 1)
    else:
        credentials = None
        host = netloc

    if ':' in host and not (host[0] == '[' and host[-1] == ']'):
        # there *is* port
        host, port = host.rsplit(':', 1)
        if not port or port.isdigit():
            return url
    else:
        port = None

    if host in ('salsa.debian.org', 'github.com'):
        if '/' not in path[1:] and port:
            path = '%s/%s' % (port, path.lstrip('/'))
        netloc = host
        if ":" in netloc:
            netloc = "[%s]" % netloc
        if (credentials is not None and
            not (credentials == 'git' and
                 scheme not in ('git', 'http', 'https'))):
            netloc = '%s@%s' % (credentials, netloc)
        if host == 'salsa.debian.org':
            scheme = 'https'
        if host == 'salsa.debian.org' and path.startswith('/cgit/'):
            path = path[5:]
        new_url = urlutils.urlparse.urlunparse(
            (scheme, netloc, path, params, query, fragment))
        if url != new_url:
            warning('Fixing up URL: %s -> %s', url, new_url)
        return new_url
    return url


def vcs_git_url_to_bzr_url(url):
    """Convert a Vcs-Git string to a Breezy URL."""

    # some packages seem to use [PATH] behind the URL to
    # indicate a subdirectory inside of the versioned tree.
    # (this is not documented in policy)

    m = re.finditer(r' \[([^] ]+)\]', url)
    try:
        m = next(m)
        url = url[:m.start()] + url[m.end():]
        subpath = m.group(1)
    except StopIteration:
        subpath = None

    from breezy.git.urls import git_url_to_bzr_url
    if ' -b ' in url:
        (url, branch) = url.split(' -b ', 1)
    else:
        branch = None

    url = fixup_broken_git_url(url)
    url = git_url_to_bzr_url(url)
    if branch:
        branch = urlutils.quote(branch, '')
        url = urlutils.join_segment_parameters(
            url, {'branch': branch})
    if subpath:
        url = urlutils.join(url, subpath)
    return url


def vcs_bzr_url_to_bzr_url(url):
    return directories.dereference(url)


def vcs_darcs_url_to_bzr_url(url):
    return url


def vcs_mtn_url_to_bzr_url(url):
    return url


def vcs_arch_url_to_bzr_url(url):
    return url


def vcs_cvs_url_to_bzr_url(location):
    from breezy.location import cvs_to_url
    try:
        (loc, module) = location.split(' ', 1)
    except ValueError:
        loc = location
        module = None
    url = cvs_to_url(loc)
    if module is not None:
        url = url + '?module=' + urlutils.quote(module)
    return url


def vcs_hg_url_to_bzr_url(url):
    if ' -b ' in url:
        (url, branch) = url.split(' -b ', 1)
    else:
        branch = None

    if branch:
        branch = urlutils.quote(branch, '')
        url = urlutils.join_segment_parameters(
            url, {'branch': branch})
    return url


def vcs_svn_url_to_bzr_url(url):
    return url


vcs_field_to_bzr_url_converters = [
    ("Bzr", vcs_bzr_url_to_bzr_url),
    ("Darcs", vcs_darcs_url_to_bzr_url),
    ("Svn", vcs_svn_url_to_bzr_url),
    ("Git", vcs_git_url_to_bzr_url),
    ("Hg", vcs_hg_url_to_bzr_url),
    ("Cvs", vcs_cvs_url_to_bzr_url),
    ("Mtn", vcs_mtn_url_to_bzr_url),
    ("Arch", vcs_arch_url_to_bzr_url),
]


def source_package_vcs_url(control):
    """Extract a Breezy-compatible URL from a source package.
    """
    (vcs_type, vcs_url) = source_package_vcs(control)
    return vcs_type, dict(vcs_field_to_bzr_url_converters)[vcs_type](vcs_url)


class AptDirectory(object):
    """Simple Bazaar directory service which uses dpkg Vcs-* fields."""

    def look_up(self, name, url):
        if "/" in name:
            (name, version) = name.split("/", 1)
        else:
            version = None

        try:
            import apt_pkg
        except ImportError as e:
            raise DependencyNotPresent('apt_pkg', e)
        apt_pkg.init()

        sources = apt_pkg.SourceRecords()

        by_version = {}
        while sources.lookup(name):
            by_version[sources.version] = sources.record

        if len(by_version) == 0:
            raise urlutils.InvalidURL(path=url, extra='package not found')

        if version is None:
            # Try the latest version
            version = sorted(by_version, key=Version)[-1]

        if version not in by_version:
            raise urlutils.InvalidURL(
                path=url, extra='version %s not found' % version)

        control = Deb822(by_version[version])

        try:
            vcs, url = source_package_vcs_url(control)
        except KeyError:
            note("Retrieving Vcs locating from %s Debian version %s", name,
                 version)
            raise urlutils.InvalidURL(path=url, extra='no VCS URL found')

        note("Resolved package URL from Debian package %s/%s: %s",
             name, version, url)
        return url


class DgitDirectory(object):
    """Directory that looks up the URL according to a Dgit control field."""

    def look_up(self, name, url):
        if "/" in name:
            (name, version) = name.split("/", 1)
        else:
            version = None

        try:
            import apt_pkg
        except ImportError as e:
            raise DependencyNotPresent('apt_pkg', e)

        apt_pkg.init()

        sources = apt_pkg.SourceRecords()

        urls = {}
        while sources.lookup(name):
            control = Deb822(sources.record)
            pkg_version = control["Version"]
            try:
                urls[pkg_version] = control["Dgit"].split(' ')
            except KeyError:
                pass

        if len(urls) == 0:
            raise urlutils.InvalidURL(path=url, extra='no URLs found')

        if version is None:
            # Try the latest version
            version = sorted(urls, key=Version)[-1]

        if version not in urls:
            raise urlutils.InvalidURL(
                path=url, extra='version %s not found' % version)

        if len(urls[version]) < 3:
            raise urlutils.InvalidURL(
                path=url,
                extra='dgit header does not have location information')

        url = urlutils.join_segment_parameters(
                urls[version][3],
                {"tag": urlutils.quote(urls[version][2], '')})

        note("Resolved package URL from Debian package %s/%s: %s",
             name, version, url)
        return url


class VcsDirectory(object):
    """Use local Vcs Directory."""

    def look_up(self, name, url):
        from debian.deb822 import Deb822
        with open('debian/control', 'r') as f:
            source = Deb822(f)
            vcs, url = source_package_vcs_url(source)
            return url


def upstream_branch_alias(b):
    from .util import debuild_config
    with b.lock_read():
        tree = b.basis_tree()
        config = debuild_config(tree)
        return directories.dereference(config.upstream_branch)



#    test_revspec.py -- Test the revision specs
#    Copyright (C) 2008 Canonical Ltd.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from __future__ import absolute_import

import os

try:
    from ....revisionspec import InvalidRevisionSpec
except ImportError:   # brz < 3.2
    from ....errors import InvalidRevisionSpec
from ....revisionspec import RevisionSpec

from ....tests.test_revisionspec import TestRevisionSpec

from . import Version, Changelog
from ..revspec import (
    UnknownVersion,
    VersionNotSpecified,
    RevisionSpec_package,
    RevisionSpec_upstream,
    )


class TestRevisionSpec_package(TestRevisionSpec):

    def test_from_string_package(self):
        spec = RevisionSpec.from_string('package:0.1-1')
        self.assertIsInstance(spec, RevisionSpec_package)
        self.assertEqual(spec.spec, '0.1-1')

    def test_simple_package(self):
        self.tree.branch.tags.set_tag('0.1-1', b'r1')
        self.assertInHistoryIs(1, b'r1', 'package:0.1-1')

    def test_unkown_version(self):
        self.assertRaises(
            UnknownVersion, self.get_in_history, 'package:0.1-1')

    def test_missing_version(self):
        self.assertRaises(
            VersionNotSpecified, self.get_in_history, 'package:')


class TestRevisionSpec_upstream(TestRevisionSpec):

    package_name = 'test'
    package_version = Version('0.1-1')
    upstream_version = property(
        lambda self: self.package_version.upstream_version)

    def make_changelog(self, version=None):
        if version is None:
            version = self.package_version
        c = Changelog()
        c.new_block()
        c.version = Version(version)
        c.package = self.package_name
        c.distributions = 'unstable'
        c.urgency = 'low'
        c.author = 'James Westby <jw+debian@jameswestby.net>'
        c.date = 'Thu,  3 Aug 2006 19:16:22 +0100'
        c.add_change('')
        c.add_change('  *  test build')
        c.add_change('')
        return c

    def write_changelog(self, changelog, filename):
        f = open(filename, 'w')
        changelog.write_to_open_file(f)
        f.close()

    def add_changelog(self, tree, version):
        cl = self.make_changelog("1.2-1")
        tree.mkdir('debian')
        self.write_changelog(
            cl, os.path.join(tree.basedir, 'debian/changelog'))
        tree.add(['debian', 'debian/changelog'])

    def test_from_string_package(self):
        self.make_branch_and_tree('.')
        spec = RevisionSpec.from_string('upstream:')
        self.assertIsInstance(spec, RevisionSpec_upstream)
        self.assertEqual(spec.spec, '')

    def test_no_changelog(self):
        t = self.make_branch_and_tree('.')
        spec = RevisionSpec.from_string('upstream:')
        self.assertRaises(InvalidRevisionSpec, spec.as_revision_id, t.branch)

    def test_version_specified(self):
        t = self.make_branch_and_tree('.')
        upstream_revid = t.commit('The upstream revision')
        t.branch.tags.set_tag("upstream-1.2", upstream_revid)
        t.commit('Mention upstream.')
        self.add_changelog(t, "1.2-1")
        spec = RevisionSpec.from_string('upstream:1.2')
        self.assertEquals(upstream_revid, spec.as_revision_id(t.branch))
        spec = RevisionSpec.from_string('upstream:1.2-1')
        self.assertEquals(upstream_revid, spec.as_revision_id(t.branch))

    def test_version_from_changelog(self):
        t = self.make_branch_and_tree('.')
        upstream_revid = t.commit('The upstream revision')
        t.branch.tags.set_tag("upstream-1.2", upstream_revid)
        t.commit('Mention upstream.')
        self.add_changelog(t, "1.2-1")
        spec = RevisionSpec.from_string('upstream:')
        self.assertEquals(upstream_revid, spec.as_revision_id(t.branch))



#    test_builder.py -- Testsuite for builddeb builder.py
#    Copyright (C) 2007 James Westby <jw+debian@jameswestby.net>
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from __future__ import absolute_import

import os

from ....tests import TestCaseInTempDir

from ..builder import (
    DebBuild,
    BuildFailedError,
    NoSourceDirError,
    )


class TestDebBuild(TestCaseInTempDir):

    def test_prepare_makes_parents(self):
        builder = DebBuild(None, 'target/sub/sub2', None)
        builder.prepare()
        self.assertPathExists('target/sub')
        self.assertPathDoesNotExist('target/sub/sub2')

    def test_prepare_purges_dir(self):
        self.build_tree(['target/', 'target/sub/'])
        builder = DebBuild(None, 'target/sub/', None)
        builder.prepare()
        self.assertPathExists('target')
        self.assertPathDoesNotExist('target/sub')

    def test_use_existing_preserves(self):
        self.build_tree(['target/', 'target/sub/'])
        builder = DebBuild(None, 'target/sub/', None, use_existing=True)
        builder.prepare()
        self.assertPathExists('target/sub')

    def test_use_existing_errors_if_not_present(self):
        self.build_tree(['target/'])
        builder = DebBuild(None, 'target/sub/', None, use_existing=True)
        self.assertRaises(NoSourceDirError, builder.prepare)
        self.assertPathDoesNotExist('target/sub')

    def test_export(self):
        class MkdirDistiller(object):
            def distill(self, target):
                os.mkdir(target)
        builder = DebBuild(MkdirDistiller(), 'target', None)
        builder.export()
        self.assertPathExists('target')

    def test_build(self):
        builder = DebBuild(None, 'target', "touch built")
        self.build_tree(['target/'])
        builder.build()
        self.assertPathExists('target/built')

    def test_build_fails(self):
        builder = DebBuild(None, 'target', "false")
        self.build_tree(['target/'])
        self.assertRaises(BuildFailedError, builder.build)

    def test_clean(self):
        builder = DebBuild(None, 'target', None)
        self.build_tree(['target/', 'target/foo'])
        builder.clean()
        self.assertPathDoesNotExist('target')



#    tags.py -- Providers of upstream source - tag names
#    Copyright (C) 2016-2020 Jelmer Vernooij <jelmer@debian.org>
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from itertools import islice
import re

from ....errors import (
    BzrError,
    )


from debmutate.versions import mangle_version_for_git


class GbpTagFormatError(BzrError):
    _fmt = 'Unknown variable %(variable)s in tag %(tag_name)s.'

    def __init__(self, tag_name, variable):
        self.variable = variable
        self.tag_name = tag_name


def gbp_expand_tag_name(tag_format, version):
    version = str(version)
    # See gbp/pkg/pkgpolicy.py in gbp-buildpackage
    version_mangle_re = (
        r'%\(version'
        r'%(?P<M>[^%])'
        r'%(?P<R>([^%]|\\%))+'
        r'\)s')

    ret = tag_format
    m = re.search(version_mangle_re, tag_format)
    if m:
        ret = re.sub(version_mangle_re, "%(version)s", tag_format)
        version = version.replace(
            m.group('M'), m.group('R').replace(r'\%', '%'))

    vars = {
        'version': version,
        'hversion': version.replace('.', '-'),
        }
    try:
        return ret % vars
    except KeyError as e:
        raise GbpTagFormatError(tag_format, e.args[0])


def upstream_tag_name(version, component=None, distro=None, git_style=False):
    if git_style:
        # In git, the convention is to use a slash
        if distro is None:
            name = "upstream/" + mangle_version_for_git(version)
        else:
            name = "upstream-%s/%s" % (distro, version.replace('~', '_'))
    else:
        if distro is None:
            name = "upstream-" + version
        else:
            name = "upstream-%s-%s" % (distro, version)
    if component is not None:
        name += "/%s" % component
    return name


def possible_upstream_tag_names(package, version, component=None):
    tags = []
    if component is None:
        # compatibility with git-buildpackage
        tags.append("upstream/%s" % version)
        tags.append("upstream-%s" % version)
        manipulated = 'upstream/%s' % mangle_version_for_git(version)
        if manipulated not in tags:
            tags.append(manipulated)
        # compatibility with svn-buildpackage
        tags.append("upstream_%s" % version)

        # common upstream names
        tags.append("%s" % version)
        tags.append("v%s" % version)
        if '~' not in str(version) and '+' not in str(version):
            tags.append("release-%s" % version)
            tags.append("v%s-release" % version)
        tags.append("%s-%s" % (package, version))
    else:
        tags.append(upstream_tag_name(version, component))
    return tags


def is_upstream_tag(tag):
    """Return true if tag is an upstream tag.

    :param tag: The string name of the tag.
    :return: True if the tag name is one generated by upstream tag operations.
    """
    return (tag.startswith('upstream-') or tag.startswith('upstream/') or
            tag.startswith('upstream_'))


def upstream_tag_version(tag):
    """Return the upstream version portion of an upstream tag name.

    :param tag: The string name of the tag.
    :return: tuple with version portion of the tag and component name
    """
    assert is_upstream_tag(tag), "Not an upstream tag: %s" % tag
    if tag.startswith('upstream/'):
        tag = tag[len('upstream/'):]
    elif tag.startswith('upstream_'):
        tag = tag[len('upstream_'):]
    elif tag.startswith('upstream-'):
        tag = tag[len('upstream-'):]
        if tag.startswith('debian-'):
            tag = tag[len('debian-'):]
        elif tag.startswith('ubuntu-'):
            tag = tag[len('ubuntu-'):]
    tag = tag.replace('_', '~')
    if '/' not in tag:
        return (None, tag)
    (version, component) = tag.rsplit('/', 1)
    if component == "":
        component = None
    return (component, version)


def _rev_is_upstream_import(revision, package, version):
    possible_messages = [
        'Import %s_%s' % (package, version),
        'import %s_%s' % (package, version),
        'import %s-%s' % (package.replace('-', '_'), version),
        'Imported upstream version %s' % version,
        'Import upstream version %s' % version,
        'New upstream version %s' % version,
        'New upstream version v%s' % version,
        '%s-%s' % (package, version),
        ]
    for possible_message in possible_messages:
        if revision.message.lower().startswith(possible_message.lower()):
            return True
    return False


def _rev_is_upstream_merge(revision, package, version):
    if revision.message.lower().startswith(
            ("Merge tag 'v%s' into debian/" % version).lower()):
        return True
    if revision.message.lower().startswith(
            ("Merge tag '%s-%s' into " % (package, version)).lower()):
        return True
    return False


def upstream_version_tag_start_revids(tag_dict, package, version):
    candidate_tag_start = [
        'debian/%s-' % mangle_version_for_git(version),
        'debian-%s' % version,
        'debian-%s-%s' % (package, version),
        # Epochs are sometimes replaced by underscores, rather than by %,
        # as DEP-14 suggests.
        'debian/%s-' % mangle_version_for_git(version.replace(':', '_')),
        ]
    for tag_name, revid in tag_dict.items():
        if any([tag_name.startswith(tag_start)
                for tag_start in candidate_tag_start]):
            yield (tag_name, revid)


def search_for_upstream_version(
        repository, start_revids, package, version, component=None, md5=None,
        scan_depth=None):
    """Find possible upstream revisions that don't have appropriate tags."""
    todo = []
    graph = repository.get_graph()
    for revid, parents in islice(
            graph.iter_ancestry(start_revids), scan_depth):
        todo.append(revid)
    for revid, rev in repository.iter_revisions(todo):
        if rev is None:
            continue
        if _rev_is_upstream_import(rev, package, version):
            return revid

    # Try again, but this time search for merge revisions
    for revid, rev in repository.iter_revisions(todo):
        if rev is None:
            continue
        if _rev_is_upstream_merge(rev, package, version):
            return rev.parent_ids[1]
    return None



#    merge_package.py -- The plugin for bzr
#    Copyright (C) 2009 Canonical Ltd.
#
#    :Author: Muharem Hrnjadovic <muharem@ubuntu.com>
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from __future__ import absolute_import

import os
import tempfile

from debian.changelog import Version

from .errors import (
    BzrError,
    MultipleUpstreamTarballsNotSupported,
    )
from .import_dsc import DistributionBranch
from .util import find_changelog


class SharedUpstreamConflictsWithTargetPackaging(BzrError):

    _fmt = ('The upstream branches for the merge source and target have '
            'diverged. Unfortunately, the attempt to fix this problem '
            'resulted in conflicts. Please resolve these, commit and '
            're-run the "%(cmd)s" command to finish. '
            'Alternatively, until you commit you can use "bzr revert" to '
            'restore the state of the unmerged branch.')

    def __init__(self, cmd):
        self.cmd = cmd


def _upstream_version_data(branch, revid):
    """Most recent upstream versions/revision IDs of the merge source/target.

    Please note: both packaging branches must have been read-locked
    beforehand.

    :param branch: The merge branch.
    :param revid: The revision in the branch to consider
    :param tree: Optional tree for the revision
    """
    db = DistributionBranch(branch, branch)
    tree = branch.repository.revision_tree(revid)
    changelog, _ignore = find_changelog(tree, '', False)
    uver = changelog.version.upstream_version
    upstream_revids = db.pristine_upstream_source.version_as_revisions(
        None, uver)
    if list(upstream_revids.keys()) != [None]:
        raise MultipleUpstreamTarballsNotSupported()
    upstream_revid = upstream_revids[None]
    return (Version(uver), upstream_revid)


def fix_ancestry_as_needed(tree, source, source_revid=None):
    r"""Manipulate the merge target's ancestry to avoid upstream conflicts.

    Merging J->I given the following ancestry tree is likely to result in
    upstream merge conflicts:

    debian-upstream                 ,------------------H
                       A-----------B                    \
    ubuntu-upstream     \           \`-------G           \
                         \           \        \           \
    debian-packaging      \ ,---------D--------\-----------J
                           C           \        \
    ubuntu-packaging        `----E------F--------I

    Here there was a new upstream release (G) that Ubuntu packaged (I), and
    then another one that Debian packaged, skipping G, at H and J.

    Now, the way to solve this is to introduce the missing link.

    debian-upstream                 ,------------------H------.
                       A-----------B                    \      \
    ubuntu-upstream     \           \`-------G-----------\------K
                         \           \        \           \
    debian-packaging      \ ,---------D--------\-----------J
                           C           \        \
    ubuntu-packaging        `----E------F--------I

    at K, which isn't a real merge, as we just use the tree from H, but add
    G as a parent and then we merge that in to Ubuntu.

    debian-upstream                 ,------------------H------.
                       A-----------B                    \      \
    ubuntu-upstream     \           \`-------G-----------\------K
                         \           \        \           \      \
    debian-packaging      \ ,---------D--------\-----------J      \
                           C           \        \                  \
    ubuntu-packaging        `----E------F--------I------------------L

    At this point we can merge J->L to merge the Debian and Ubuntu changes.

    :param tree: The `WorkingTree` of the merge target branch.
    :param source: The merge source (packaging) branch.
    """
    upstreams_diverged = False
    t_upstream_reverted = False
    target = tree.branch

    with source.lock_read():
        if source_revid is None:
            source_revid = source.last_revision()
        with tree.lock_write():
            # "Unpack" the upstream versions and revision ids for the merge
            # source and target branch respectively.
            (us_ver, us_revid) = _upstream_version_data(source, source_revid)
            (ut_ver, ut_revid) = _upstream_version_data(
                    target, target.last_revision())

            # Did the upstream branches of the merge source/target diverge?
            graph = source.repository.get_graph(target.repository)
            upstreams_diverged = (len(graph.heads([us_revid, ut_revid])) > 1)

            # No, we're done!
            if not upstreams_diverged:
                return (upstreams_diverged, t_upstream_reverted)

            # Instantiate a `DistributionBranch` object for the merge target
            # (packaging) branch.
            db = DistributionBranch(tree.branch, tree.branch)
            with tempfile.TemporaryDirectory(dir=os.path.join(tree.basedir, '..')) as tempdir:
                # Extract the merge target's upstream tree into a temporary
                # directory.
                db.extract_upstream_tree({None: ut_revid}, tempdir)
                tmp_target_utree = db.pristine_upstream_tree

                # Merge upstream branch tips to obtain a shared upstream
                # parent.  This will add revision K (see graph above) to a
                # temporary merge target upstream tree.
                with tmp_target_utree.lock_write():
                    if us_ver > ut_ver:
                        # The source upstream tree is more recent and the
                        # temporary target tree needs to be reshaped to match
                        # it.
                        tmp_target_utree.revert(
                            None, source.repository.revision_tree(us_revid))
                        t_upstream_reverted = True

                    tmp_target_utree.set_parent_ids((ut_revid, us_revid))
                    new_revid = tmp_target_utree.commit(
                        'Prepared upstream tree for merging into target '
                        'branch.')

                    # Repository updates during a held lock are not visible,
                    # hence the call to refresh the data in the /target/ repo.
                    tree.branch.repository.refresh_data()

                    tree.branch.fetch(source, us_revid)
                    tree.branch.fetch(tmp_target_utree.branch, new_revid)

                    # Merge shared upstream parent into the target merge
                    # branch. This creates revison L in the digram above.
                    conflicts = tree.merge_from_branch(tmp_target_utree.branch)
                    if conflicts > 0:
                        cmd = "bzr merge"
                        raise SharedUpstreamConflictsWithTargetPackaging(cmd)
                    else:
                        tree.commit(
                            'Merging shared upstream rev into target branch.')

    return (upstreams_diverged, t_upstream_reverted)



#    test_directory.py -- Testsuite for builddeb directory.py
#    Copyright (C) 2019 Jelmer Vernooij <jelmer@jelmer.uk>
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from __future__ import absolute_import

from ....tests import TestCase

from ..directory import (
    fixup_broken_git_url,
    vcs_git_url_to_bzr_url,
    vcs_hg_url_to_bzr_url,
    vcs_cvs_url_to_bzr_url,
    )


class VcsGitUrlToBzrUrlTests(TestCase):

    def test_preserves(self):
        self.assertEqual(
            'git://github.com/jelmer/dulwich',
            vcs_git_url_to_bzr_url('git://github.com/jelmer/dulwich'))
        self.assertEqual(
            'https://github.com/jelmer/dulwich',
            vcs_git_url_to_bzr_url('https://github.com/jelmer/dulwich'))

    def test_with_branch(self):
        self.assertEqual(
            'https://github.com/jelmer/dulwich,branch=foo',
            vcs_git_url_to_bzr_url('https://github.com/jelmer/dulwich -b foo'))

    def test_with_subpath(self):
        self.assertEqual(
            'https://github.com/jelmer/dulwich/path',
            vcs_git_url_to_bzr_url(
                'https://github.com/jelmer/dulwich [path]'))
        self.assertEqual(
            'https://github.com/jelmer/dulwich,branch=foo/path',
            vcs_git_url_to_bzr_url(
                'https://github.com/jelmer/dulwich -b foo [path]'))

    def test_fixup(self):
        self.assertEqual(
            'git://github.com/jelmer/dulwich',
            vcs_git_url_to_bzr_url('git://github.com:jelmer/dulwich'))


class FixUpGitUrlTests(TestCase):

    def test_salsa_not_https(self):
        self.assertEqual(
            'https://salsa.debian.org/jelmer/dulwich',
            fixup_broken_git_url(
                'git://salsa.debian.org/jelmer/dulwich'))

    def test_salsa_uses_cgit(self):
        self.assertEqual(
            'https://salsa.debian.org/jelmer/dulwich',
            fixup_broken_git_url(
                'https://salsa.debian.org/cgit/jelmer/dulwich'))


class VcsHgUrlToBzrUrlTests(TestCase):

    def test_preserves(self):
        self.assertEqual(
            'https://bitbucket.org/jelmer/dulwich',
            vcs_hg_url_to_bzr_url('https://bitbucket.org/jelmer/dulwich'))

    def test_with_branch(self):
        self.assertEqual(
            'https://bitbucket.org/jelmer/dulwich,branch=foo',
            vcs_hg_url_to_bzr_url('https://bitbucket.org/jelmer/dulwich -b foo'))


class VcsCvsUrlToBzrUrlTests(TestCase):

    def setUp(self):
        super(VcsCvsUrlToBzrUrlTests, self).setUp()
        import breezy
        if breezy.version_info < (3, 1, 1):
            self.skipTest('version of breezy too old')

    def test_pserver(self):
        self.assertEqual(
            'cvs+pserver://anonymous@cvs.savannah.nongnu.org'
            '/cvsroot/fkt?module=debian/unstable',
            vcs_cvs_url_to_bzr_url(
                ':pserver:anonymous@cvs.savannah.nongnu.org:'
                '/cvsroot/fkt debian/unstable'))
        self.assertEqual(
            'cvs+pserver://anonymous@cvs.savannah.nongnu.org'
            '/cvsroot/fkt',
            vcs_cvs_url_to_bzr_url(
                ':pserver:anonymous@cvs.savannah.nongnu.org:'
                '/cvsroot/fkt'))



#    merge_upstream.py -- Merge new upstream versions of packages.
#    Copyright (C) 2007 Reinhard Tartler <siretart@tauware.de>
#                  2007 James Westby <jw+debian@jameswestby.net>
#                  2008-2019 Jelmer Vernooij <jelmer@samba.org>
#
#    Code is also taken from bzrtools, which is
#             (C) 2005, 2006, 2007 Aaron Bentley <aaron.bentley@utoronto.ca>
#             (C) 2005, 2006 Canonical Limited.
#             (C) 2006 Michael Ellerman.
#    and distributed under the GPL, version 2 or later.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from __future__ import absolute_import

import os
import subprocess
import tempfile

from debian.changelog import Version

from ... import osutils
from ...revision import NULL_REVISION

from .errors import (
    BzrError,
    )
from .import_dsc import (
    DistributionBranch,
    DistributionBranchSet,
    PreviousVersionTagMissing,
    UpstreamAlreadyImported,
    )
from .repack_tarball import repack_tarball
from .util import (
    component_from_orig_tarball,
    tarball_name,
    FORMAT_3_0_QUILT,
    FORMAT_3_0_NATIVE,
    tree_get_source_format,
    )


TAG_PREFIX = "upstream-"


class DchError(BzrError):
    _fmt = 'There was an error using dch: %(error)s.'

    def __init__(self, error):
        BzrError.__init__(self, error=error)


def package_version(upstream_version, distribution_name, epoch=None):
    """Determine the package version for a new upstream.

    :param upstream_version: Upstream version string
    :param distribution_name: Distribution the package is for
    :param epoch: Optional epoch
    """
    if distribution_name == "ubuntu":
        ret = Version("%s-0ubuntu1" % upstream_version)
    else:
        ret = Version("%s-1" % upstream_version)
    ret.epoch = epoch
    return ret


def upstream_merge_changelog_line(upstream_version):
    """Describe that a new upstream revision was merged.

    This will either describe that a new upstream release or a new upstream
    snapshot was merged.

    :param upstream_version: Upstream version string
    :return: Line string for use in changelog
    """
    vcs_suffixes = ["~bzr", "+bzr", "~svn", "+svn", "~git", "+git", "-git"]
    for vcs_suffix in vcs_suffixes:
        if vcs_suffix in str(upstream_version):
            entry_description = "New upstream snapshot."
            break
    else:
        entry_description = "New upstream release."
    return entry_description


# TODO(jelmer): Move into debmutate
def changelog_add_new_version(
        tree, subpath, upstream_version, distribution_name, changelog,
        package):
    """Add an entry to the changelog for a new version.

    :param tree: WorkingTree in which the package lives
    :param upstream_version: Upstream version to add
    :param distribution_name: Distribution name (debian, ubuntu, ...)
    :param changelog: Changelog object
    :param package: Package name
    """
    entry_description = upstream_merge_changelog_line(upstream_version)
    if changelog is None:
        epoch = None
    else:
        epoch = changelog.epoch
    argv = ["dch", "-v",
            str(package_version(upstream_version, distribution_name, epoch)),
            "-D", "UNRELEASED", "--release-heuristic", "changelog",
            "--package", package, entry_description]
    if not tree.has_filename(osutils.pathjoin(subpath, 'debian')):
        tree.mkdir(osutils.pathjoin(subpath, 'debian'))
    cl_path = osutils.pathjoin(subpath, "debian/changelog")
    create = (not tree.has_filename(cl_path))
    if create:
        argv.append("--create")
    proc = subprocess.Popen(
        argv, cwd=osutils.pathjoin(tree.basedir, subpath),
        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (stdout, stderr) = proc.communicate()
    if proc.returncode != 0:
        raise DchError("Adding changelog entry failed: %s" % stderr)
    if not tree.is_versioned(cl_path):
        tree.add([cl_path])


def do_import(
        tree, subpath, tarball_filenames, package, version,
        current_version, upstream_branch, upstream_revisions, merge_type=None,
        force=False, force_pristine_tar=False, committer=None,
        files_excluded=None):
    """Import new tarballs.

    Args:
      tree: Working tree to operate in
      tarball_filenames: List of tarball filenames as tuples with (path,
        component)
      package: Package name
      version: New upstream version to merge
      current_version: Current upstream version in tree
      upstream_branch: Optional upstream branch to merge from
      upstream_revisions: Dictionary mapping versions to upstream revisions
      merge_type: Merge type
      committer: Committer string to use
      files_excluded: Files to exclude
    """
    db = DistributionBranch(tree.branch, tree.branch, tree=tree)
    dbs = DistributionBranchSet()
    dbs.add_branch(db)
    tarballs = [(p, component_from_orig_tarball(p, package, version), None)
                for p in tarball_filenames]

    with tempfile.TemporaryDirectory(
            dir=os.path.join(db.tree.basedir, '..')) as tempdir:
        if current_version is not None:
            db._export_previous_upstream_tree(
                package, current_version, tempdir)
        else:
            db.create_empty_upstream_tree(tempdir)
        if db.pristine_upstream_source.has_version(package, version):
            raise UpstreamAlreadyImported(version)

        parents = {None: []}
        if db.pristine_upstream_branch.last_revision() != NULL_REVISION:
            parents = {None: [db.pristine_upstream_branch.last_revision()]}

        return db.import_upstream_tarballs(
            tarballs, package, version,
            parents,
            upstream_branch=upstream_branch,
            upstream_revisions=upstream_revisions,
            force_pristine_tar=force_pristine_tar,
            committer=committer, files_excluded=files_excluded)


def do_merge(
        tree, subpath, tarball_filenames, package, version,
        current_version, upstream_branch, upstream_revisions, merge_type=None,
        force=False, force_pristine_tar=False, committer=None,
        files_excluded=None):
    """Actually execute a merge.

    Args:
      tree: Working tree to operate in
      tarball_filenames: List of tarball filenames as tuples with (path,
        component)
      package: Package name
      version: New upstream version to merge
      current_version: Current upstream version in tree
      upstream_branch: Optional upstream branch to merge from
      upstream_revisions: Dictionary mapping versions to upstream revisions
      merge_type: Merge type
      committer: Committer string to use
      files_excluded: Files to exclude
    """
    db = DistributionBranch(tree.branch, tree.branch, tree=tree)
    dbs = DistributionBranchSet()
    dbs.add_branch(db)
    tarballs = [(p, component_from_orig_tarball(p, package, version))
                for p in tarball_filenames]
    return db.merge_upstream(
        tarballs, package, version, current_version,
        upstream_branch=upstream_branch,
        upstream_revisions=upstream_revisions,
        merge_type=merge_type, force=force,
        force_pristine_tar=force_pristine_tar,
        committer=committer, files_excluded=files_excluded)


def fetch_tarball(package, version, orig_dir, locations, v3):
    ret = []
    format = None
    for location in locations:
        if v3:
            if location.endswith(".tar.bz2") or location.endswith(".tbz2"):
                format = "bz2"
            elif location.endswith(".tar.xz"):
                format = "xz"
        dest_name = tarball_name(package, version, None, format=format)
        tarball_filename = os.path.join(orig_dir, dest_name)
        repack_tarball(location, dest_name, target_dir=orig_dir)
        ret.append(tarball_filename)
    return ret


def get_tarballs(orig_dir, tree, package, version, locations):
    """Retrieve upstream tarballs."""
    source_format = tree_get_source_format(tree)
    v3 = (source_format in [FORMAT_3_0_QUILT, FORMAT_3_0_NATIVE])
    orig_dir = os.path.join(tree.basedir, orig_dir)
    if not os.path.exists(orig_dir):
        os.makedirs(orig_dir)
    return fetch_tarball(package, version, orig_dir, locations, v3)



#    test_config.py -- Tests for builddeb's config.py
#    Copyright (C) 2007 James Westby <jw+debian@jameswestby.net>
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from __future__ import absolute_import

from ....branch import Branch

from ..config import (
    BUILD_TYPE_MERGE,
    DebBuildConfig,
    UpstreamMetadataSyntaxError,
    )
from . import TestCaseWithTransport


class DebBuildConfigTests(TestCaseWithTransport):

    def setUp(self):
        super(DebBuildConfigTests, self).setUp()
        self.tree = self.make_branch_and_tree('.')
        self.branch = self.tree.branch
        with open('default.conf', 'w') as f:
            f.write('['+DebBuildConfig.section+']\n')
            # shouldn't be read as it needs to be trusted
            f.write('builder = invalid builder\n')
            f.write('build-dir = default build dir\n')
            f.write('orig-dir = default orig dir\n')
            f.write('result-dir = default result dir\n')
        with open('user.conf', 'w') as f:
            f.write('['+DebBuildConfig.section+']\n')
            f.write('builder = valid builder\n')
            f.write('quick-builder = valid quick builder\n')
            f.write('orig-dir = user orig dir\n')
            f.write('result-dir = user result dir\n')
        with open('.bzr/branch/branch.conf', 'w') as f:
            f.write('['+DebBuildConfig.section+']\n')
            f.write('quick-builder = invalid quick builder\n')
            f.write('result-dir = branch result dir\n')
        self.tree.add(['default.conf', 'user.conf'])
        self.config = DebBuildConfig(
            [('user.conf', True), ('default.conf', False)],
            branch=self.branch)

    def test_secure_not_from_untrusted(self):
        self.assertEqual(self.config.builder, 'valid builder')

    def test_secure_not_from_branch(self):
        self.assertEqual(self.config.quick_builder, 'valid quick builder')

    def test_branch_over_all(self):
        self.assertEqual(self.config.result_dir, 'branch result dir')

    def test_hierarchy(self):
        self.assertEqual(self.config.orig_dir, 'user orig dir')
        self.assertEqual(self.config.build_dir, 'default build dir')

    def test_no_entry(self):
        self.assertEqual(self.config.merge, False)
        self.assertEqual(self.config.build_type, None)

    def test_parse_error(self):
        with open('invalid.conf', 'w') as f:
            f.write('['+DebBuildConfig.section+'\n')
        DebBuildConfig([('invalid.conf', True, 'invalid.conf')])

    def test_upstream_metadata(self):
        cfg = DebBuildConfig([], tree=self.branch.basis_tree())
        self.assertIs(None, cfg.upstream_branch)

        self.build_tree_contents([
          ('debian/',),
          ('debian/upstream/',),
          ('debian/upstream/metadata',
           b'Name: example\n'
           b'Repository: http://example.com/foo\n'
           b'Repository-Tag-Prefix: exampl-\n'
           )])
        self.tree.add(
            ['debian', 'debian/upstream', 'debian/upstream/metadata'])

        cfg = DebBuildConfig([], tree=self.tree)
        self.assertEquals("http://example.com/foo", cfg.upstream_branch)
        self.assertEquals(
            "tag:exampl-$UPSTREAM_VERSION", cfg.export_upstream_revision)

    def test_invalid_upstream_metadata(self):
        cfg = DebBuildConfig([], tree=self.branch.basis_tree())
        self.assertIs(None, cfg.upstream_branch)

        self.build_tree_contents([
          ('debian/',),
          ('debian/upstream/',),
          ('debian/upstream/metadata',
           b'debian/changelog.blah'
           )])
        self.tree.add(
            ['debian', 'debian/upstream', 'debian/upstream/metadata'])

        self.assertRaises(
            UpstreamMetadataSyntaxError, DebBuildConfig, [], tree=self.tree)

try:
    from ...svn.config import SubversionBuildPackageConfig  # noqa: F401
except ImportError:
    pass
else:
    from ...svn.tests import SubversionTestCase

    class DebuildSvnBpTests(SubversionTestCase):

        if not getattr(SubversionTestCase, "make_svn_branch", None):
            def make_svn_branch(self, relpath):
                repos_url = self.make_repository(relpath)
                return Branch.open(repos_url)

        def test_from_properties(self):
            branch = self.make_svn_branch("d")

            cfg = DebBuildConfig([], tree=branch.basis_tree())
            self.assertEquals(False, cfg.merge)

            dc = self.get_commit_editor(branch.base)
            d = dc.add_dir("debian")
            d.change_prop("mergeWithUpstream", "1")
            d.change_prop("svn-bp:origDir", "someorigdir")
            dc.close()

            cfg = DebBuildConfig([], tree=branch.basis_tree())
            self.assertEquals(True, cfg.merge)
            self.assertEquals(BUILD_TYPE_MERGE, cfg.build_type)
            self.assertEquals("someorigdir", cfg.orig_dir)

        def test_from_svn_layout_file(self):
            branch = self.make_svn_branch("d")

            cfg = DebBuildConfig([], tree=branch.basis_tree())
            self.assertEquals(False, cfg.merge)

            with self.get_commit_editor(branch.base) as dc:
                d = dc.add_dir("debian")
                f = d.add_file("debian/svn-layout")
                f.modify(b"origDir = someorigdir\n")

            cfg = DebBuildConfig([], tree=branch.basis_tree())
            self.assertEquals("someorigdir", cfg.orig_dir)



#    upstream.py -- Providers of upstream source
#    Copyright (C) 2009 Canonical Ltd.
#    Copyright (C) 2009-2020 Jelmer Vernooij <jelmer@debian.org>
#
#    This file is part of brz-debian.
#
#    brz-debian is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    brz-debian is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with brz-debian; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import re
import subprocess
import shutil
import sys
import tempfile

from debmutate.watch import parse_watch_file

from ....errors import BzrError, NoSuchFile
from .... import osutils
from ....export import export
from ....trace import note, warning
from . import UpstreamSource, PackageVersionNotPresent


class UScanError(BzrError):

    _fmt = "UScan failed to run: %(errors)s."

    def __init__(self, errors):
        self.errors = errors


class NoWatchFile(BzrError):

    _fmt = "Tree %(tree)r has no watch file %(path)s."

    def __init__(self, tree, path):
        BzrError.__init__(self, tree=tree, path=path)


class WatchLineWithoutMatches(BzrError):

    _fmt = "No matching files for watch line %(line)r."

    def __init__(self, line):
        BzrError.__init__(self, line=line)


class WatchLineWithoutMatchingHrefs(PackageVersionNotPresent):

    _fmt = ("No match for %(mangled_version)s was not found "
            "for %(line)r in %(upstream)s.")

    def __init__(self, mangled_version, line, package, version, upstream):
        BzrError.__init__(self, package=package, version=version,
                          upstream=upstream, mangled_version=mangled_version,
                          line=line)


class UScanSource(UpstreamSource):
    """Upstream source that uses uscan."""

    def __init__(self, tree, subpath=None, top_level=False):
        self.tree = tree
        self.subpath = subpath
        self.top_level = top_level

    @classmethod
    def from_tree(cls, tree, subpath, top_level=False):
        if top_level:
            file = 'watch'
        else:
            file = 'debian/watch'
        if subpath:
            file = osutils.pathjoin(subpath, file)
        if not tree.has_filename(file):
            raise NoWatchFile(tree, file)
        return cls(tree, subpath=subpath, top_level=top_level)

    def _export_file(self, name, directory):
        if self.top_level:
            file = name
        else:
            file = 'debian/' + name
        if self.subpath:
            file = osutils.pathjoin(self.subpath, file)
        if not self.tree.has_filename(file):
            raise NoSuchFile(file, self.tree)
        output_path = os.path.join(directory, 'debian', name)
        output_dir = os.path.dirname(output_path)
        if not os.path.isdir(output_dir):
            os.makedirs(output_dir)
        with open(output_path, 'wb') as f:
            f.write(self.tree.get_file_text(file))
        return output_path

    def get_latest_version(self, package, current_version):
        with tempfile.TemporaryDirectory() as tmpdir:
            try:
                watch_tempfilename = self._export_file('watch', tmpdir)
            except NoSuchFile:
                note("No watch file to use to check latest upstream release.")
                return None
            args = ["--watchfile=%s" % watch_tempfilename,
                    "--package=%s" % package, "--report",
                    "--no-download",
                    "--upstream-version=%s" % current_version]
            text, retcode = _run_dehs_uscan(args, cwd=tmpdir)
            uversionmangle = None
            with open(watch_tempfilename, 'r') as f:
                wf = parse_watch_file(f)
                if wf and wf.entries and len(wf.entries) == 1:
                    uversionmangle = getattr(
                        wf.entries[0], 'uversionmangle', None)
        version = _xml_report_extract_upstream_version(text)
        if version is None:
            for w in _xml_report_extract_warnings(text):
                if re.match(
                        'In (.*)/watch no matching files for watch line',
                        w.splitlines()[0]):
                    raise WatchLineWithoutMatches(w.splitlines()[1])
                raise UScanError(w)
            return
        if uversionmangle is None:
            return version
        return uversionmangle(version)

    def get_recent_versions(self, package, since_version=None):
        raise NotImplementedError(self.get_recent_versions)

    def fetch_tarballs(self, package, version, target_dir, components=None):
        note("Using uscan to look for the upstream tarball.")
        with tempfile.TemporaryDirectory() as tmpdir:
            container = os.path.join(tmpdir, 'container')
            os.mkdir(container)
            if self.top_level:
                subdir = ''
            else:
                subdir = 'debian'
            if self.subpath:
                subdir = osutils.pathjoin(self.subpath, subdir)
            # Just export all of debian/, since e.g. uupdate needs more of it.
            export(self.tree, os.path.join(container, 'debian'), format='dir',
                   subdir=subdir)
            args = ["--force-download", "--rename",
                    "--check-dirname-level=0",
                    "--download", '--destdir=%s' % container,
                    "--download-debversion=%s" % version]
            text, r = _run_dehs_uscan(args, cwd=container)
            _xml_report_extract_errors(text)
            orig_files = _xml_report_extract_target_paths(text)
            if not orig_files:
                for w in _xml_report_extract_warnings(text):
                    m = re.match(
                        'In (.*) no matching hrefs for version (.*) in watch '
                        'line', w.splitlines()[0])
                    if m:
                        raise WatchLineWithoutMatchingHrefs(
                            m.group(1), w.splitlines()[1],
                            package, version, self)
                    raise UScanError(w)
                raise PackageVersionNotPresent(package, version, self)
            _xml_report_print_warnings(text)
            if all([os.path.exists(p) for p in orig_files]):
                pass
            else:
                orig_files = [
                    entry.path for entry in os.scandir(tmpdir)
                    if entry.name != 'container']
                if not orig_files:
                    note("uscan could not find the needed tarballs.")
                    raise PackageVersionNotPresent(package, version, self)
            ret = []
            for src in orig_files:
                dst = os.path.join(target_dir, os.path.basename(src))
                ret.append(dst)
                shutil.copy(os.path.join(tmpdir, src), dst)

            return [src for src in ret if not src.endswith('.asc')]


def _xml_report_extract_upstream_version(text):
    _xml_report_extract_errors(text)
    from xml.sax.saxutils import unescape
    # uscan --dehs's output isn't well-formed XML, so let's fall back to
    # regexes instead..
    m = re.search(b'<upstream-version>(.*)</upstream-version>', text)
    if not m:
        return None
    return unescape(m.group(1).decode())


def _xml_report_extract_target_paths(text):
    from xml.sax.saxutils import unescape
    return [
        unescape(m.group(1).decode())
        for m in re.finditer(b'<target-path>(.*)</target-path>', text)]


def _run_dehs_uscan(args, cwd):
    p = subprocess.Popen(
        ["uscan", "--dehs"] + args,
        stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
    (stdout, stderr) = p.communicate()
    if b'</dehs>' not in stdout:
        error = stderr.decode()
        for line in error.splitlines():
            if line.startswith('uscan warn: '):
                continue
            if line.startswith('uscan error '):
                line = line[len('uscan error '):]
            raise UScanError(line)
        raise UScanError(error)
    sys.stderr.write(stderr.decode())
    return stdout, p.returncode


def _xml_report_extract_warnings(text):
    from xml.sax.saxutils import unescape
    for m in re.finditer(
            b"<warnings>(.*?)</warnings>", text,
            flags=(re.M | re.S)):
        yield unescape(m.group(1).decode())


def _xml_report_print_warnings(text):
    for w in _xml_report_extract_warnings(text):
        warning(w)


def _xml_report_extract_errors(text):
    from xml.sax.saxutils import unescape
    lines = [unescape(m.group(1).decode())
             for m in re.finditer(
                 b"<errors>(.*?)</errors>", text,
                 flags=(re.M | re.S))]
    ignored = []
    for line in lines:
        if not line.startswith('uscan warn: '):
            raise UScanError(line)
        else:
            ignored.append(line[len('uscan warn: '):])

    for line in lines:
        raise UScanError(line)

    return ignored
